import os
import platform
import time
from datetime import datetime, timedelta

import pytest

import picologging
from picologging.handlers import (
    RotatingFileHandler,
    TimedRotatingFileHandler,
    WatchedFileHandler,
)


def test_filehandler(tmp_path):
    log_file = tmp_path / "log.txt"
    handler = picologging.FileHandler(log_file)
    logger = picologging.getLogger("test")
    logger.setLevel(picologging.DEBUG)
    logger.addHandler(handler)
    logger.warning("test")
    handler.close()

    with open(log_file) as f:
        assert f.read() == "test\n"


def test_filehandler_delay(tmp_path):
    log_file = tmp_path / "log.txt"
    handler = picologging.FileHandler(log_file, delay=True)
    logger = picologging.getLogger("test")
    logger.setLevel(picologging.DEBUG)
    logger.addHandler(handler)
    logger.warning("test")
    handler.close()

    with open(log_file) as f:
        assert f.read() == "test\n"


@pytest.mark.skipif(platform.system() == "Windows", reason="Not supported on Windows.")
def test_watchedfilehandler(tmp_path):
    log_file = tmp_path / "log.txt"
    handler = WatchedFileHandler(log_file)
    logger = picologging.getLogger("test")
    logger.setLevel(picologging.DEBUG)
    logger.addHandler(handler)
    logger.warning("test")
    handler.close()

    with open(log_file) as f:
        assert f.read() == "test\n"


@pytest.mark.skipif(platform.system() == "Windows", reason="Not supported on Windows.")
def test_watchedfilehandler_file_changed(tmp_path):
    log_file = tmp_path / "log.txt"
    handler = WatchedFileHandler(log_file)
    logger = picologging.getLogger("test")
    logger.setLevel(picologging.DEBUG)
    logger.addHandler(handler)

    os.remove(log_file)
    with open(log_file, "w"):
        ...

    logger.warning("test")
    handler.close()

    with open(log_file) as f:
        assert f.read() == "test\n"


@pytest.mark.skipif(platform.system() == "Windows", reason="Not supported on Windows.")
def test_watchedfilehandler_file_removed(tmp_path):
    log_file = tmp_path / "log.txt"
    handler = WatchedFileHandler(log_file)
    logger = picologging.getLogger("test")
    logger.setLevel(picologging.DEBUG)
    logger.addHandler(handler)

    os.remove(log_file)

    logger.warning("test")
    handler.close()

    with open(log_file) as f:
        assert f.read() == "test\n"


def test_rotatingfilehandler(tmp_path):
    log_file = tmp_path / "log.txt"
    handler = RotatingFileHandler(log_file, maxBytes=1, backupCount=2)
    logger = picologging.getLogger("test")
    logger.setLevel(picologging.DEBUG)
    logger.addHandler(handler)

    for _ in range(5):
        logger.warning("test")
    handler.close()

    with open(log_file) as f:
        assert f.read() == "test\n"

    for i in range(1, 3):
        log_file = tmp_path / f"log.txt.{i}"

        with open(log_file) as f:
            assert f.read() == "test\n"


def test_rotatingfilehandler_avoids_non_regular_files(tmp_path, monkeypatch):
    log_file = tmp_path / "log.txt"
    handler = RotatingFileHandler(log_file, maxBytes=1, backupCount=2)
    logger = picologging.getLogger("test")
    logger.setLevel(picologging.DEBUG)
    logger.addHandler(handler)

    logger.warning("test")

    monkeypatch.setattr(os.path, "isfile", lambda _: False)
    logger.warning("test")
    handler.close()


def test_rotatingfilehandler_without_maxbytes(tmp_path):
    log_file = tmp_path / "log.txt"
    handler = RotatingFileHandler(log_file)
    logger = picologging.getLogger("test")
    logger.setLevel(picologging.DEBUG)
    logger.addHandler(handler)

    logger.warning("test")
    handler.close()


def test_baserotatinghandler_callable_rotator(tmp_path):
    log_file = tmp_path / "log.txt"
    handler = RotatingFileHandler(log_file, maxBytes=1, backupCount=1)
    handler.rotator = lambda src, dst: os.rename(src, dst)
    logger = picologging.getLogger("test")
    logger.setLevel(picologging.DEBUG)
    logger.addHandler(handler)

    logger.warning("test")
    logger.warning("test")
    handler.close()
    assert sorted(os.listdir(tmp_path)) == ["log.txt", "log.txt.1"]


def test_baserotatinghandler_callable_namer(tmp_path):
    log_file = tmp_path / "log.txt"
    handler = RotatingFileHandler(log_file, maxBytes=1, backupCount=1)
    handler.namer = lambda name: name + ".5"
    logger = picologging.getLogger("test")
    logger.setLevel(picologging.DEBUG)
    logger.addHandler(handler)

    logger.warning("test")
    logger.warning("test")
    handler.close()
    assert sorted(os.listdir(tmp_path)) == ["log.txt", "log.txt.1.5"]


def test_filehandler_repr(tmp_path):
    log_file = tmp_path / "log.txt"
    handler = picologging.FileHandler(log_file)
    assert repr(handler) == f"<FileHandler {log_file} (NOTSET)>"


@pytest.mark.parametrize("utc", [False, True])
def test_timed_rotatingfilehandler_rollover(tmp_path, utc):
    log_file = tmp_path / "log.txt"
    handler = TimedRotatingFileHandler(log_file, when="S", backupCount=1, utc=utc)
    logger = picologging.getLogger("test")
    logger.setLevel(picologging.DEBUG)
    logger.addHandler(handler)

    logger.warning("test")
    handler.rollover_at = time.time() - 1

    logger.warning("test")
    handler.close()

    files = os.listdir(tmp_path)
    assert len(files) == 2

    for file_name in files:
        with open(tmp_path / file_name) as file:
            assert file.read() == "test\n"


def test_timed_rotatingfilehandler_rollover_removes_old_files(tmp_path):
    with open(tmp_path / "log.txt.1970-11-01_00-00-00", "w"):
        ...

    log_file = tmp_path / "log.txt"
    handler = TimedRotatingFileHandler(log_file, when="S", backupCount=1)
    logger = picologging.getLogger("test")
    logger.setLevel(picologging.DEBUG)
    logger.addHandler(handler)

    logger.warning("test")
    handler.rollover_at = time.time() - 1
    logger.warning("test")
    handler.close()

    assert len(os.listdir(tmp_path)) == 2


def test_timed_rotatingfilehandler_rollover_keeps_non_related_files(tmp_path):
    with open(tmp_path / "normal_file.txt", "w"):
        ...

    log_file = tmp_path / "log.txt"
    handler = TimedRotatingFileHandler(log_file, when="S", backupCount=1)
    logger = picologging.getLogger("test")
    logger.setLevel(picologging.DEBUG)
    logger.addHandler(handler)

    logger.warning("test")
    handler.rollover_at = time.time() - 1
    logger.warning("test")
    handler.close()

    assert len(os.listdir(tmp_path)) == 3


def test_timed_rotatingfilehandler_rollover_removes_existing_log(tmp_path, monkeypatch):
    existing_log_file = tmp_path / "log.txt.2022-07-01_16-00-00"
    with open(existing_log_file, "w"):
        ...

    log_file = tmp_path / "log.txt"
    handler = TimedRotatingFileHandler(log_file, when="S", backupCount=1)
    logger = picologging.getLogger("test")
    logger.setLevel(picologging.DEBUG)
    logger.addHandler(handler)

    logger.warning("test")

    monkeypatch.setattr(handler, "rotation_filename", lambda _: existing_log_file)
    handler.rollover_at = time.time() - 1
    logger.warning("test")
    handler.close()

    assert len(os.listdir(tmp_path)) == 2


def test_timed_rotatingfilehandler_non_existing_file(tmp_path, monkeypatch):
    monkeypatch.setattr(os.path, "exists", lambda _: False)

    log_file = tmp_path / "log.txt"
    handler = TimedRotatingFileHandler(log_file, when="S", backupCount=1)
    logger = picologging.getLogger("test")
    logger.setLevel(picologging.DEBUG)
    logger.addHandler(handler)
    handler.close()


def test_timed_rotatingfilehandler_when_intervals(tmp_path):
    when_interval = [
        ("S", 1),
        ("M", 60),
        ("H", 60 * 60),
        ("D", 60 * 60 * 24),
        ("MIDNIGHT", 60 * 60 * 24),
        ("W1", 60 * 60 * 24 * 7),
    ]

    log_file = tmp_path / "log.txt"
    for when, interval in when_interval:
        handler = TimedRotatingFileHandler(log_file, when=when)
        assert handler.interval == interval
        handler.close()


def test_timed_rotatingfilehandler_invalid_when(tmp_path):
    log_file = tmp_path / "log.txt"

    with pytest.raises(ValueError):
        TimedRotatingFileHandler(log_file, when="W")

    with pytest.raises(ValueError):
        TimedRotatingFileHandler(log_file, when="W7")

    with pytest.raises(ValueError):
        TimedRotatingFileHandler(log_file, when="X")


def test_timed_rotatingfilehandler_utc(tmp_path):
    log_file = tmp_path / "log.txt"
    handler = TimedRotatingFileHandler(log_file, when="MIDNIGHT", utc=True)
    handler.close()
    assert handler.rollover_at > time.time()


def test_timed_rotatingfilehandler_atime(tmp_path):
    log_file = tmp_path / "log.txt"
    at_time = (datetime.now() - timedelta(hours=1)).time()
    handler = TimedRotatingFileHandler(log_file, when="MIDNIGHT", atTime=at_time)
    handler.close()
    assert handler.rollover_at > time.time()


def test_timed_rotatingfilehandler_avoids_non_regular_files(tmp_path, monkeypatch):
    log_file = tmp_path / "log.txt"
    handler = TimedRotatingFileHandler(log_file, backupCount=1)
    logger = picologging.getLogger("test")
    logger.setLevel(picologging.DEBUG)
    logger.addHandler(handler)

    logger.warning("test")

    monkeypatch.setattr(os.path, "isfile", lambda _: False)
    logger.warning("test")
    handler.close()



import datetime
import logging
import sys

import pytest

from picologging import Formatter, LogRecord


def test_formatter_default_fmt():
    f = Formatter()
    assert f.datefmt is None
    record = LogRecord(
        "hello", logging.WARNING, __file__, 123, "bork bork bork", (), None
    )
    s = f.format(record)
    assert s == "bork bork bork"


def test_formatter_custom_fmt():
    f = Formatter("%(name)s %(levelname)s %(message)s")
    assert f.datefmt is None
    record = LogRecord(
        "hello", logging.WARNING, __file__, 123, "bork bork bork", (), None
    )
    s = f.format(record)
    assert s == "hello WARNING bork bork bork"
    assert f.usesTime() == False
    assert f.formatMessage(record) == s


def test_formatter_default_fmt_against_builtin():
    pico_f = Formatter()
    logging_f = logging.Formatter()

    assert pico_f.datefmt == logging_f.datefmt

    pico_record = LogRecord(
        "hello", logging.WARNING, __file__, 123, "bork bork bork", (), None
    )
    logging_record = logging.LogRecord(
        "hello", logging.WARNING, "/serv/", 123, "bork bork bork", (), None
    )
    assert pico_f.format(pico_record) == logging_f.format(logging_record)
    assert pico_f.format(pico_record) == logging_f.format(pico_record)
    assert pico_f.formatMessage(pico_record) == logging_f.formatMessage(pico_record)


def test_formatter_default_fmt_exc_info_against_builtin():
    pico_f = Formatter()
    logging_f = logging.Formatter()

    try:
        raise Exception("error")
    except Exception:
        exc_info = sys.exc_info()

    pico_record = LogRecord(
        "hello", logging.WARNING, __file__, 123, "bork bork bork", (), exc_info, "", ""
    )
    logging_record = logging.LogRecord(
        "hello", logging.WARNING, "/serv/", 123, "bork bork bork", (), exc_info, "", ""
    )
    assert pico_f.format(pico_record) == logging_f.format(logging_record)
    assert pico_f.format(pico_record) == logging_f.format(pico_record)
    assert pico_f.formatMessage(pico_record) == logging_f.formatMessage(pico_record)


def test_formatter_custom_datefmt():
    f = Formatter("%(name)s %(levelname)s %(message)s", datefmt="%Y-%m-%d")
    assert f.datefmt == "%Y-%m-%d"
    record = LogRecord(
        "hello", logging.WARNING, __file__, 123, "bork bork bork", (), None
    )
    s = f.format(record)
    assert s == "hello WARNING bork bork bork"


def test_formatter_explicit_none_datefmt_style():
    f = Formatter("%(name)s %(levelname)s %(message)s", None, "%")
    assert f.datefmt is None
    record = LogRecord(
        "hello", logging.WARNING, __file__, 123, "bork bork bork", (), None
    )
    s = f.format(record)
    assert s == "hello WARNING bork bork bork"


possible_format_strings = [
    "%(message)s",
    "%(message)s is a potato",
    "%(name)s",
    "%(msg)s",
    "%(args)s",
    "%(levelname)s",
    "%(levelno)s",
    "%(pathname)s",
    "%(filename)s",
    "%(module)s",
    "%(funcName)s",
    "%(lineno)d",
    "%(threadName)s",
    "%(process)d",
    "%(processName)s",
    "%(stack_info)s",
    "%(exc_info)s",
    "%(exc_text)s",
]


@pytest.mark.parametrize("field", possible_format_strings)
def test_format_field(field):
    pico_f = Formatter(field)
    log_f = logging.Formatter(field)
    record = LogRecord(
        "hello", logging.WARNING, __file__, 123, "bork bork bork", (), None
    )
    assert pico_f.format(record) == log_f.format(record)


def test_format_time():
    pico_f = Formatter("%(msecs)d %(relativeCreated)d")
    record = LogRecord(
        "hello", logging.WARNING, __file__, 123, "bork bork bork", (), None
    )
    assert pico_f.format(record)


def test_asctime_field():
    pico_f = Formatter("%(asctime)s")
    record = LogRecord(
        "hello", logging.WARNING, __file__, 123, "bork bork bork", (), None
    )
    assert pico_f.format(record)
    assert pico_f.usesTime()


def test_record_with_stack_info():
    pico_f = Formatter("%(message)s")
    record = LogRecord(
        "hello",
        logging.WARNING,
        __file__,
        123,
        "bork bork bork",
        (),
        None,
        None,
        "hello",
    )
    assert pico_f.format(record) == "bork bork bork\nhello"


def test_record_with_non_str_sstack_info():
    pico_f = Formatter("%(message)s")
    record = LogRecord(
        "hello",
        logging.WARNING,
        __file__,
        123,
        "bork bork bork",
        (),
        None,
        None,
        ["hello", "world"],
    )
    assert pico_f.format(record) == "bork bork bork\n['hello', 'world']"


def test_format_stack():
    pico_f = Formatter("%(message)s")
    assert pico_f.formatStack([1, 2, 3]) == [1, 2, 3]


def test_delete_formatter():
    pico_f = Formatter("%(message)s")
    del pico_f


def test_formatter_bad_init_args():
    with pytest.raises(TypeError):
        Formatter(dog="good boy")


def test_formatter_bad_style():
    with pytest.raises(ValueError):
        Formatter(style="!")


def test_formatter_bad_style_type():
    with pytest.raises(TypeError):
        Formatter(style=123)


def test_formatter_bad_fmt_type():
    record = LogRecord(
        "hello", logging.WARNING, __file__, 123, "bork bork bork", (), None
    )
    with pytest.raises((TypeError, ValueError)):
        f = Formatter(fmt=123)
        assert f.format(record) == "bork bork bork"


def test_formatter_with_validate_flag_and_invalid_fmt():
    f = Formatter(fmt="%(message ", validate=True)
    record = LogRecord(
        "hello", logging.WARNING, __file__, 123, "bork bork bork", (), None
    )
    assert f.format(record) == "%(message "


def test_datefmt_bad_type():
    with pytest.raises(TypeError):
        Formatter(datefmt=123)


def test_format_with_custom_datefmt():
    actual_date = datetime.datetime.now().strftime("%Y-%m-%d")
    f = Formatter("%(name)s %(levelname)s %(message)s %(asctime)s", datefmt="%Y-%m-%d")
    assert f.datefmt == "%Y-%m-%d"
    record = LogRecord(
        "hello", logging.WARNING, __file__, 123, "bork bork bork", (), None
    )
    s = f.format(record)
    assert s == f"hello WARNING bork bork bork {actual_date}"
    assert f.usesTime() == True
    assert f.formatMessage(record) == s


def test_formatter_repr():
    f = Formatter("%(message)s")
    assert repr(f) == "<Formatter: fmt='%(message)s'>"


def test_exc_info_invalid_type():
    record = LogRecord(
        "hello", logging.WARNING, __file__, 123, "bork bork bork", (), (1, 2, 3)
    )
    f = Formatter()
    with pytest.raises(AttributeError):
        f.format(record)


def test_exc_info_invalid_value_types():
    record = LogRecord(
        "hello", logging.WARNING, __file__, 123, "bork bork bork", (), [1, 2, 3]
    )
    f = Formatter()
    with pytest.raises(TypeError):
        f.format(record)


# TODO #41 : test defaults are propagating to string formatters


def test_formatter_templates():
    # Not supported, so check it doesn't just crash
    with pytest.raises(NotImplementedError):
        Formatter("%(message)s", style="$")


def test_format_exception():
    pico_f = Formatter("%(message)s")

    try:
        raise Exception("error")
    except Exception:
        ei = sys.exc_info()

    result = pico_f.formatException(ei)
    assert result.startswith("Traceback (most recent call last):")
    assert result.endswith(
        'test_format_exception\n    raise Exception("error")\nException: error'
    )



import sys

from picologging import Formatter


def test_format_exception():
    pico_f = Formatter("%(message)s")

    try:
        raise Exception("error")
    except Exception:
        ei = sys.exc_info()

    result = pico_f.formatException(ei)
    assert result.startswith("Traceback (most recent call last):")
    assert result.endswith(
        'test_format_exception\n    raise Exception("error")\nException: error'
    )


if __name__ == "__main__":
    for _ in range(100_000):
        test_format_exception()



import io
import threading

import picologging
from picologging import Logger, StreamHandler


def test_threaded_execution():
    logger = Logger("test", picologging.DEBUG)
    tmp = io.StringIO()
    handler = StreamHandler(tmp)
    logger.addHandler(handler)

    def _log_message():
        logger.debug("from thread")

    t = threading.Thread(target=_log_message)
    t.start()
    t.join()
    result = tmp.getvalue()
    assert result == "from thread\n"



from io import StringIO

import picologging


def log(level=picologging.INFO):
    logger = picologging.Logger("test", level)
    tmp = StringIO()

    handler = picologging.StreamHandler(tmp)
    handler.setLevel(level)
    formatter = picologging.Formatter(
        "%(created)f/%(asctime)s %(pathname)s:%(module)s:%(filename)s:%(lineno)d %(funcName)s %(levelno)d %(name)s - %(levelname)s %(process)d %(thread)d- %(message)s"
    )
    handler.setFormatter(formatter)
    logger.handlers.append(handler)

    for _ in range(100_000):
        logger.debug("There has been a picologging issue")
        logger.debug("There has been a picologging issue %s %s %s", 1, 2, 3)
        logger.info("There has been a picologging issue %s %s %s", 1, 2, 3)
        logger.warning("There has been a picologging issue %s %s %s", 1, 2, 3)

    assert len(tmp.getvalue()) > 100_000


if __name__ == "__main__":
    log()



import io
import logging
import uuid

import pytest

import picologging


def test_logger_attributes():
    logger = picologging.Logger("test")
    assert logger.name == "test"
    assert logger.level == logging.NOTSET
    assert logger.parent is None
    assert logger.propagate == True
    assert logger.handlers == []
    assert logger.disabled == False
    assert logger.propagate == True


level_names = [
    "DEBUG",
    "INFO",
    "WARNING",
    "ERROR",
    "CRITICAL",
    "NOTSET",
]

levels = [
    logging.DEBUG,
    logging.INFO,
    logging.WARNING,
    logging.ERROR,
    logging.CRITICAL,
    logging.NOTSET,
]


@pytest.mark.parametrize("level", levels)
def test_logging_custom_level(level):
    logger = picologging.Logger("test", level)
    assert logger.level == level


def test_custom_logger_has_no_parent():
    logger = picologging.Logger("test")
    assert logger.parent is None


def test_remove_non_existent_handler():
    logger = picologging.Logger("test")
    assert logger.removeHandler("handler") is None


def test_set_level():
    logger = picologging.Logger("test")
    logger.setLevel(logging.DEBUG)
    assert logger.level == logging.DEBUG


def test_disabled_logger():
    logger = picologging.Logger("test", logging.DEBUG)
    logger.disabled = True
    stream = io.StringIO()
    logger.handlers.append(picologging.StreamHandler(stream))
    assert logger.debug("Hello World") is None
    result = stream.getvalue()
    assert result == ""
    ex = Exception("arghhh!!")
    logger.exception("Hello World", ex)
    result = stream.getvalue()
    assert result == ""


def test_logger_with_logging_handler():
    logger = picologging.Logger("test", logging.DEBUG)
    stream = io.StringIO()
    handler = logging.StreamHandler(stream)
    handler.setFormatter(logging.Formatter("%(message)s"))
    logger.addHandler(handler)
    assert logger.debug("Hello World") is None
    result = stream.getvalue()
    assert result == "Hello World\n"


def test_get_effective_level():
    logger = picologging.Logger("test")
    parent = picologging.Logger("parent", logging.DEBUG)
    logger.parent = parent
    assert logger.getEffectiveLevel() == logging.DEBUG
    assert logger.level == logging.NOTSET
    logger.setLevel(logging.WARNING)
    assert logger.getEffectiveLevel() == logging.WARNING


def test_dodgy_parents():
    logger = picologging.Logger("test")
    parent = "potato"
    with pytest.raises(TypeError):
        logger.parent = parent

    with pytest.raises(TypeError):
        logger.parent = logging.getLogger("test")
        logger.getEffectiveLevel()

    with pytest.raises(TypeError):
        del logger.parent


def test_add_filter():
    logger = picologging.Logger("test")
    filter = picologging.Filter("filter1")
    logger.addFilter(filter)
    assert logger.filters == [filter]
    filter2 = picologging.Filter("filter2")
    logger.addFilter(filter2)
    assert logger.filters == [filter, filter2]


def test_remove_filter():
    logger = picologging.Logger("test")
    filter = picologging.Filter("filter1")
    logger.addFilter(filter)
    assert logger.filters == [filter]
    logger.removeFilter(filter)
    assert logger.filters == []
    logger.removeFilter(filter)
    assert logger.filters == []


def test_delete_filter():
    filter = picologging.Filter("filter1")
    del filter


def test_filterer_direct_type():
    filterable = picologging.Filterer()
    assert filterable.filters == []
    filter = picologging.Filter("filter1")
    filterable.addFilter(filter)
    assert filterable.filters == [filter]
    filter2 = picologging.Filter("filter2")
    filterable.addFilter(filter2)
    assert filterable.filters == [filter, filter2]
    filterable.removeFilter(filter)
    assert filterable.filters == [filter2]
    filterable.removeFilter(filter2)
    assert filterable.filters == []
    del filterable


def test_no_filter():
    logger = picologging.Logger("test")
    record = picologging.LogRecord("test", logging.INFO, "test", 1, "test", (), {})
    assert logger.filter(record) == True


def test_filter_record():
    logger = picologging.Logger("test")
    filter = picologging.Filter("hello")
    logger.addFilter(filter)
    record = picologging.LogRecord("hello", logging.INFO, "test", 1, "test", (), {})
    record2 = picologging.LogRecord("goodbye", logging.INFO, "test", 1, "test", (), {})
    assert logger.filter(record) == True
    assert logger.filter(record2) == False
    logger.removeFilter(filter)
    assert logger.filter(record) == True
    assert logger.filter(record2) == True


def test_filter_callable():
    logger = picologging.Logger("test")

    def filter(record):
        return record.name == "hello"

    logger.addFilter(filter)
    record = picologging.LogRecord("hello", logging.INFO, "test", 1, "test", (), {})
    assert logger.filter(record) == True
    record = picologging.LogRecord("goodbye", logging.INFO, "test", 1, "test", (), {})
    assert logger.filter(record) == False


def test_log_debug():
    logger = picologging.Logger("test", logging.DEBUG)
    stream = io.StringIO()
    handler = picologging.StreamHandler(stream)
    handler.setFormatter(picologging.Formatter("%(message)s"))
    logger.addHandler(handler)
    assert logger.debug("Hello World") is None
    result = stream.getvalue()
    assert result == "Hello World\n"


def test_log_debug_info_level_logger():
    logger = picologging.Logger("test", logging.INFO)
    stream = io.StringIO()
    logger.handlers.append(picologging.StreamHandler(stream))
    assert logger.debug("Hello World") is None
    result = stream.getvalue()
    assert result == ""


def test_log_debug_info_level_logger_logging_handler():
    logger = picologging.Logger("test", logging.INFO)
    stream = io.StringIO()
    logger.handlers.append(logging.StreamHandler(stream))
    assert logger.debug("Hello World") is None
    result = stream.getvalue()
    assert result == ""


@pytest.mark.parametrize("level", levels)
def test_log_log(level):
    logger = picologging.Logger("test", level)
    stream = io.StringIO()
    handler = picologging.StreamHandler(stream)
    handler.setFormatter(picologging.Formatter("%(message)s"))
    logger.addHandler(handler)
    assert logger.log(level, "Hello World") is None
    result = stream.getvalue()
    assert result == "Hello World\n"


def test_logger_with_explicit_level(capsys):
    logger = picologging.Logger("test", logging.DEBUG)
    tmp = io.StringIO()
    handler = picologging.StreamHandler(tmp)
    handler.setLevel(logging.DEBUG)
    formatter = picologging.Formatter("%(name)s - %(levelname)s - %(message)s")
    handler.setFormatter(formatter)
    logger.handlers.append(handler)
    logger.debug("There has been a picologging issue")
    result = tmp.getvalue()
    assert result == "test - DEBUG - There has been a picologging issue\n"
    cap = capsys.readouterr()
    assert cap.out == ""
    assert cap.err == ""


def test_exception_capture():
    logger = picologging.Logger("test", logging.DEBUG)
    tmp = io.StringIO()
    handler = picologging.StreamHandler(tmp)
    logger.addHandler(handler)
    try:
        1 / 0
    except ZeroDivisionError:
        logger.exception("bork")
    result = tmp.getvalue()
    assert "bork" in result
    assert "ZeroDivisionError: division by zero" in result


def test_getlogger_no_args():
    logger = logging.getLogger()
    assert logger.name == "root"
    assert logger.level == logging.WARNING
    assert logger.parent is None

    picologger = picologging.getLogger()
    assert picologger.name == "root"
    assert picologger.level == logging.WARNING
    assert picologger.parent is None


def test_logger_init_bad_args():
    with pytest.raises(TypeError):
        logger = picologging.Logger("goo", 10, dog=1)

    with pytest.raises(TypeError):
        logger = picologging.Logger(name="test", level="potato")


@pytest.mark.parametrize("level", levels)
def test_logger_repr(level):
    logger = picologging.Logger("test", level)
    assert repr(logger) == f"<Logger 'test' ({level_names[levels.index(level)]})>"


def test_logger_repr_effective_level():
    logger = picologging.Logger("test")
    logger.parent = picologging.Logger("parent", picologging.WARNING)
    assert repr(logger) == f"<Logger 'test' (WARNING)>"


def test_logger_repr_invalid_level():
    logger = picologging.Logger("test", level=100)
    assert repr(logger) == f"<Logger 'test' ()>"


def test_set_level_bad_type():
    logger = picologging.Logger("goo", picologging.DEBUG)
    with pytest.raises(TypeError):
        logger.setLevel("potato")


def test_add_remove_handlers():
    logger = picologging.Logger("goo", picologging.DEBUG)
    assert logger.handlers == []
    test_handler = picologging.Handler()
    logger.addHandler(test_handler)
    # add it twice should have no effect
    logger.addHandler(test_handler)
    assert len(logger.handlers) == 1
    assert test_handler in logger.handlers
    logger.removeHandler(test_handler)
    assert test_handler not in logger.handlers
    assert logger.handlers == []


@pytest.mark.parametrize("level_config", levels)
def test_log_and_handle(level_config):
    logger = picologging.Logger("test", level=level_config)
    tmp = io.StringIO()
    logger.addHandler(picologging.StreamHandler(tmp))
    assert not logger.info("info_message")
    assert not logger.debug("debug_message")
    assert not logger.warning("warning_message")
    assert not logger.error("error_message")
    assert not logger.fatal("fatal_message")
    assert not logger.critical("critical_message")
    assert not logger.log(level_config, "log_message")

    tmp_value = tmp.getvalue()

    if level_config <= picologging.DEBUG and level_config != picologging.NOTSET:
        assert "debug_message" in tmp_value
    else:
        assert "debug_message" not in tmp_value
    if level_config <= picologging.INFO and level_config != picologging.NOTSET:
        assert "info_message" in tmp_value
    else:
        assert "info_message" not in tmp_value
    if level_config <= picologging.WARNING and level_config != picologging.NOTSET:
        assert "warning_message" in tmp_value
    else:
        assert "warning_message" not in tmp_value
    if level_config <= picologging.ERROR and level_config != picologging.NOTSET:
        assert "error_message" in tmp_value
    else:
        assert "error_message" not in tmp_value
    if level_config <= picologging.FATAL and level_config != picologging.NOTSET:
        assert "fatal_message" in tmp_value
    else:
        assert "fatal_message" not in tmp_value
    if level_config <= picologging.CRITICAL and level_config != picologging.NOTSET:
        assert "critical_message" in tmp_value
    else:
        assert "critical_message" not in tmp_value
    assert "log_message" in tmp_value


def test_log_xx_bad_arguments():
    logger = picologging.Logger("test", level=picologging.DEBUG)

    with pytest.raises(TypeError):
        logger.info()
    with pytest.raises(TypeError):
        logger.debug()
    with pytest.raises(TypeError):
        logger.warning()
    with pytest.raises(TypeError):
        logger.error()
    with pytest.raises(TypeError):
        logger.fatal()
    with pytest.raises(TypeError):
        logger.critical()


def test_log_bad_arguments():
    logger = picologging.Logger("test")
    with pytest.raises(TypeError):
        logger.log("potato", "message")

    with pytest.raises(TypeError):
        logger.log()


def test_notset_parent_level_match():
    logger_child = picologging.Logger("child", picologging.NOTSET)
    logger_parent = picologging.Logger("parent", picologging.DEBUG)
    logger_child.parent = logger_parent

    parent_io = io.StringIO()
    child_io = io.StringIO()
    logger_child.addHandler(picologging.StreamHandler(child_io))
    logger_parent.addHandler(picologging.StreamHandler(parent_io))

    logger_child.info("child message")
    logger_parent.info("parent message")
    parent_value = parent_io.getvalue()
    child_value = child_io.getvalue()
    assert "child message" in child_value
    assert "child message" in parent_value
    assert "parent message" in parent_value
    assert "parent message" not in child_value


def test_error_parent_level():
    logger_child = picologging.Logger("child", picologging.WARNING)
    logger_parent = picologging.Logger("parent", picologging.ERROR)
    logger_child.parent = logger_parent

    parent_io = io.StringIO()
    child_io = io.StringIO()
    logger_child.addHandler(picologging.StreamHandler(child_io))
    logger_parent.addHandler(picologging.StreamHandler(parent_io))

    logger_child.info("info message")
    logger_child.warning("warning message")
    logger_child.error("error message")

    parent_value = parent_io.getvalue()
    child_value = child_io.getvalue()

    assert "info message" not in child_value
    assert "info message" not in parent_value
    assert "warning message" in child_value
    assert "warning message" in parent_value
    assert "error message" in child_value
    assert "error message" in parent_value


def test_nested_frame_stack():
    logger = picologging.Logger("test", level=picologging.DEBUG)
    tmp = io.StringIO()
    logger.addHandler(picologging.StreamHandler(tmp))

    def f():
        def g():
            logger.info("message", stack_info=True)

        g()

    f()
    result = tmp.getvalue()
    assert "message" in result
    assert " in g\n" in result
    assert " in f\n" in result


def test_exception_object_as_exc_info():
    e = Exception("arghhh!!")
    logger = picologging.Logger("test", level=picologging.DEBUG)
    tmp = io.StringIO()
    logger.addHandler(picologging.StreamHandler(tmp))
    logger.info("message", exc_info=e)
    result = tmp.getvalue()
    assert "message" in result
    assert "arghhh!!" in result


def test_logger_setlevel_resets_other_levels():
    stream = io.StringIO()
    handler = picologging.StreamHandler(stream)
    logger = picologging.getLogger("test")
    logger.addHandler(handler)
    logger.setLevel(picologging.WARNING)

    logger.debug("test")
    assert stream.getvalue() == ""

    logger.warning("test")
    assert stream.getvalue() == "test\n"

    logger.setLevel(picologging.ERROR)
    logger.warning("test")
    assert stream.getvalue() == "test\n"

    logger.error("test")
    assert stream.getvalue() == "test\ntest\n"


def test_getlogger_root_level():
    # Check that the root logger defaults to WARNING
    logger = picologging.getLogger()
    assert logger.getEffectiveLevel() == picologging.WARNING


def test_getlogger_nonroot_levels():
    """
    Check that descendant loggers get the root level on construction.

    Before:
    | Logger       | Level     | Effective Level |
    |--------------|-----------|-----------------|
    | root         | 30        | 30              |

    After:
    | child        | 0         | 30              |
    | grandchild   | 0         | 30              |
    """
    child_name = str(uuid.uuid4())
    child_logger = picologging.getLogger(str(uuid.uuid4()))
    assert child_logger.level == picologging.NOTSET
    assert child_logger.getEffectiveLevel() == picologging.WARNING

    grandchild_name = f"{child_name}.str(uuid.uuid4())"
    grandchild_logger = picologging.getLogger(grandchild_name)
    assert grandchild_logger.level == picologging.NOTSET
    assert grandchild_logger.getEffectiveLevel() == picologging.WARNING


def test_getlogger_parentchild_levels():
    """
    Check interaction of setLevel with logger hierarchy

    Before:
    | Logger       | Level     | Effective Level |
    |--------------|-----------|-----------------|
    | root         | 30        | 30              |
    | parent       | 0         | 30              |

    After parent.setLevel(INFO):
    | parent       | 20        | 20              |

    After construction of child logger:
    | child        | 0         | 20              |

    After child.setLevel(WARNING):
    | child        | 30        | 30              |
    """
    parent_name = str(uuid.uuid4())
    parent_logger = picologging.getLogger(parent_name)
    assert parent_logger.getEffectiveLevel() == picologging.WARNING

    parent_logger.setLevel(picologging.INFO)
    assert parent_logger.getEffectiveLevel() == picologging.INFO

    child_name = f"{parent_name}.{uuid.uuid4()}"
    child_logger = picologging.getLogger(child_name)
    assert child_logger.getEffectiveLevel() == picologging.INFO

    child_logger.setLevel(picologging.WARNING)
    assert child_logger.getEffectiveLevel() == picologging.WARNING


def test_getlogger_setlevel_after():
    """
    Check for setting parent level after child construction

    Before:
    | Logger       | Level     | Effective Level |
    |--------------|-----------|-----------------|
    | root         | 30        | 30              |
    | parent       | 0         | 30              |

    After parent.setLevel(INFO):
    | parent       | 20        | 20              |

    After construction of child logger:
    | child        | 0         | 20              |

    After parent.setLevel(DEBUG):
    | parent       | 10        | 10              |
    | child        | 0         | 10              |
    """
    parent_name = str(uuid.uuid4())
    parent_logger = picologging.getLogger(parent_name)
    parent_logger.setLevel(picologging.WARNING)

    child_name = f"{parent_name}.{uuid.uuid4()}"
    child_logger = picologging.getLogger(child_name)

    parent_logger.setLevel(picologging.DEBUG)
    assert child_logger.getEffectiveLevel() == picologging.DEBUG


def test_getlogger_setlevel_after_multiple_children():
    """
    Check for setting parent level after child construction

    Before:
    | Logger       | Level     | Effective Level |
    |--------------|-----------|-----------------|
    | root         | 30        | 30              |
    | parent       | 0         | 30              |

    After parent.setLevel(WARNING):
    | parent       | 30        | 30              |

    After construction of child loggers:
    | child1       | 0         | 30              |
    | child2       | 0         | 30              |

    After parent.setLevel(DEBUG):
    | parent       | 10        | 10              |
    | child1       | 0         | 10              |
    | child2       | 0         | 10              |
    """
    parent_name = str(uuid.uuid4())
    parent_logger = picologging.getLogger(parent_name)
    parent_logger.setLevel(picologging.WARNING)

    child1_name = f"{parent_name}.{uuid.uuid4()}"
    child2_name = f"{parent_name}.{uuid.uuid4()}"
    child1_logger = picologging.getLogger(child1_name)
    child2_logger = picologging.getLogger(child2_name)

    parent_logger.setLevel(picologging.DEBUG)
    assert child1_logger.getEffectiveLevel() == picologging.DEBUG
    assert child2_logger.getEffectiveLevel() == picologging.DEBUG


def test_getlogger_setlevel_message_handled():
    """
    Check for child creation before parent creation
    and appropriate handling of messages.

    Before:
    | Logger       | Level     | Effective Level |
    |--------------|-----------|-----------------|
    | root         | 30        | 30              |
    | child        | 0         | 30              |

    After construction of parent logger:
    | parent       | 0         | 30              |
    | child        | 0         | 30              |

    After parent.setLevel(DEBUG):
    | parent       | 10        | 10              |
    | child        | 0         | 10              |
    """
    parent_name = str(uuid.uuid4())
    child_name = f"{parent_name}.{uuid.uuid4()}"
    child_logger = picologging.getLogger(child_name)
    assert child_logger.level == picologging.NOTSET
    assert child_logger.getEffectiveLevel() == picologging.WARNING

    parent_logger = picologging.getLogger(parent_name)
    stream = io.StringIO()
    handler = picologging.StreamHandler(stream)
    child_logger.addHandler(handler)

    child_logger.log(picologging.DEBUG, "Hello World")
    assert stream.getvalue() == ""

    parent_logger.setLevel(picologging.DEBUG)
    assert child_logger.getEffectiveLevel() == picologging.DEBUG
    child_logger.log(picologging.DEBUG, "Hello World")
    assert stream.getvalue() == "Hello World\n"


def test_getlogger_with_placeholder_parent():
    # Logging levels when some parent does not exist yet.
    stream = io.StringIO()
    handler = picologging.StreamHandler(stream)

    top_logger = picologging.getLogger("A")
    top_logger.addHandler(handler)
    bottom1_logger = picologging.getLogger("A.B.C")
    bottom2_logger = picologging.getLogger("A.B.D")
    top_logger.setLevel(picologging.INFO)

    assert top_logger.getEffectiveLevel() == picologging.INFO
    assert bottom1_logger.getEffectiveLevel() == picologging.INFO
    assert bottom2_logger.getEffectiveLevel() == picologging.INFO

    # These logs should be handled
    top_logger.log(picologging.WARN, "TLW")
    top_logger.info("TLI")
    bottom1_logger.log(picologging.WARN, "BL1W")
    bottom1_logger.info("BL1I")
    bottom2_logger.log(picologging.WARN, "BL2W")
    bottom2_logger.info("BL2I")

    # These should not
    top_logger.log(picologging.DEBUG, "TLD")
    bottom1_logger.debug("BLD")

    assert stream.getvalue() == "TLW\nTLI\nBL1W\nBL1I\nBL2W\nBL2I\n"

    # Now breathe life into the placeholder
    middle_logger = picologging.getLogger("A.B")
    assert middle_logger.getEffectiveLevel() == picologging.INFO



import logging
import threading

import pytest

from picologging import INFO, Formatter, LogRecord, StrFormatStyle


def test_strformatstyle():
    perc = StrFormatStyle("{msg} {levelno} {name}")
    record = LogRecord("test", INFO, __file__, 1, "hello", (), None, None, None)
    assert perc.format(record) == "hello 20 test"


def test_strformatstyle_from_formatter():
    perc = Formatter("{msg} {levelno} {name}", style="{")
    record = LogRecord("test", INFO, __file__, 1, "hello", (), None, None, None)
    assert perc.format(record) == "hello 20 test"


def test_strformatstyle_format_bad_argument():
    perc = StrFormatStyle("{msg} {levelno} {name}")
    with pytest.raises(AttributeError):
        perc.format(None)
    with pytest.raises(AttributeError):
        perc.format("")
    with pytest.raises(AttributeError):
        perc.format({})


def test_custom_attribute():
    perc = StrFormatStyle("{custom}")
    record = LogRecord("test", INFO, __file__, 1, "hello", (), None, None, None)
    record.custom = "custom"
    assert perc.format(record) == "custom"


def test_strformatstyle_bad_init_args():
    with pytest.raises(TypeError):
        StrFormatStyle(dog="good boy")


def test_funcname_format_string():
    perc = StrFormatStyle("{funcname}")
    record = LogRecord("test", INFO, __file__, 1, "hello", (), None, "superfunc", None)
    record.funcName = "superFunc"
    assert perc.format(record) == "superFunc"


def test_thread_id():
    perc = StrFormatStyle("{thread}")
    record = LogRecord("test", INFO, __file__, 1, "hello", (), None, None, None)
    assert record.thread == threading.get_ident()
    assert perc.format(record) == str(record.thread)


def test_record_created():
    perc = StrFormatStyle("{created}")
    record = LogRecord("test", INFO, __file__, 1, "hello", (), None, None, None)
    assert perc.format(record) == str(record.created)


def test_custom_field_not_an_attribute():
    perc = StrFormatStyle("{custom}")
    record = LogRecord("test", INFO, __file__, 1, "hello", (), None, None, None)
    with pytest.raises(AttributeError):
        assert perc.format(record)


def test_strformatstyle_repr():
    perc = StrFormatStyle("{msg} {levelno} {name}")
    assert repr(perc) == "<FormatStyle fmt='{msg} {levelno} {name}' style='{'>"


def test_record_with_defaults():
    perc = StrFormatStyle(
        "{msg} {levelno} {name} {fruit}", defaults={"fruit": "banana"}
    )
    assert repr(perc) == "<FormatStyle fmt='{msg} {levelno} {name} {fruit}' style='{'>"
    record = LogRecord("test", INFO, __file__, 1, "hello", (), None, None, None)
    assert perc.format(record) == "hello 20 test banana"


def test_format_logging_record():
    perc = StrFormatStyle(
        "{msg} {levelno} {name} {fruit}", defaults={"fruit": "banana"}
    )
    record = logging.LogRecord("test", INFO, __file__, 1, "hello", (), None, None, None)
    assert perc.format(record) == "hello 20 test banana"



import re
from logging.config import BaseConfigurator

import picologging
from picologging.handlers import MemoryHandler

IDENTIFIER = re.compile("^[a-z_][a-z0-9_]*$", re.I)


def valid_ident(s):
    m = IDENTIFIER.match(s)
    if not m:
        raise ValueError("Not a valid Python identifier: %r" % s)
    return True


def _resolve(name):
    """Resolve a dotted name to a global object."""
    name = name.split(".")
    used = name.pop(0)
    found = __import__(used)
    for n in name:
        used = used + "." + n
        try:
            found = getattr(found, n)
        except AttributeError:
            __import__(used)
            found = getattr(found, n)
    return found


def _handle_existing_loggers(existing, child_loggers, disable_existing):
    """
    When (re)configuring logging, handle loggers which were in the previous
    configuration but are not in the new configuration. There's no point
    deleting them as other threads may continue to hold references to them;
    and by disabling them, you stop them doing any logging.
    However, don't disable children of named loggers, as that's probably not
    what was intended by the user. Also, allow existing loggers to NOT be
    disabled if disable_existing is false.
    """
    root = picologging.root
    for log in existing:
        logger = root.manager.loggerDict[log]
        if log in child_loggers:
            logger.setLevel(picologging.NOTSET)
            logger.handlers = []
            logger.propagate = True
        else:
            logger.disabled = disable_existing


class DictConfigurator(BaseConfigurator):
    """
    Configure logging using a dictionary-like object to describe the
    configuration.
    """

    def configure(self):
        config = self.config
        if "version" not in config:
            raise ValueError("dictionary doesn't specify a version")
        if config["version"] != 1:
            raise ValueError("Unsupported version: %s" % config["version"])
        incremental = config.pop("incremental", False)
        if incremental:
            raise ValueError("Incremental option is not supported.")

        EMPTY_DICT = {}
        disable_existing = config.pop("disable_existing_loggers", True)

        # Do formatters first - they don't refer to anything else
        formatters = config.get("formatters", EMPTY_DICT)
        for name in formatters:
            try:
                formatters[name] = self.configure_formatter(formatters[name])
            except Exception as e:
                raise ValueError("Unable to configure " "formatter %r" % name) from e
        # Next, do filters - they don't refer to anything else, either
        filters = config.get("filters", EMPTY_DICT)
        for name in filters:
            try:
                filters[name] = self.configure_filter(filters[name])
            except Exception as e:
                raise ValueError("Unable to configure " "filter %r" % name) from e

        # Next, do handlers - they refer to formatters and filters
        # As handlers can refer to other handlers, sort the keys
        # to allow a deterministic order of configuration
        handlers = config.get("handlers", EMPTY_DICT)
        deferred = []
        for name in sorted(handlers):
            try:
                handler = self.configure_handler(handlers[name])
                handler.name = name
                handlers[name] = handler
            except Exception as e:
                if "target not configured yet" in str(e.__cause__):
                    deferred.append(name)
                else:
                    raise ValueError("Unable to configure handler " "%r" % name) from e

        # Now do any that were deferred
        for name in deferred:
            try:
                handler = self.configure_handler(handlers[name])
                handler.name = name
                handlers[name] = handler
            except Exception as e:
                raise ValueError("Unable to configure handler " "%r" % name) from e

        # Next, do loggers - they refer to handlers and filters

        # we don't want to lose the existing loggers,
        # since other threads may have pointers to them.
        # existing is set to contain all existing loggers,
        # and as we go through the new configuration we
        # remove any which are configured. At the end,
        # what's left in existing is the set of loggers
        # which were in the previous configuration but
        # which are not in the new configuration.
        root = picologging.root
        existing = list(root.manager.loggerDict.keys())
        # The list needs to be sorted so that we can
        # avoid disabling child loggers of explicitly
        # named loggers. With a sorted list it is easier
        # to find the child loggers.
        existing.sort()
        # We'll keep the list of existing loggers
        # which are children of named loggers here...
        child_loggers = []
        # now set up the new ones...
        loggers = config.get("loggers", EMPTY_DICT)
        for name in loggers:
            if name in existing:
                i = existing.index(name) + 1  # look after name
                prefixed = name + "."
                pflen = len(prefixed)
                num_existing = len(existing)
                while i < num_existing:
                    if existing[i][:pflen] == prefixed:
                        child_loggers.append(existing[i])
                    i += 1
                existing.remove(name)
            try:
                self.configure_logger(name, loggers[name])
            except Exception as e:
                raise ValueError("Unable to configure logger " "%r" % name) from e

        # Disable any old loggers. There's no point deleting
        # them as other threads may continue to hold references
        # and by disabling them, you stop them doing any logging.
        # However, don't disable children of named loggers, as that's
        # probably not what was intended by the user.
        _handle_existing_loggers(existing, child_loggers, disable_existing)

        # And finally, do the root logger
        root = config.get("root", None)
        if root:
            try:
                self.configure_root(root)
            except Exception as e:
                raise ValueError("Unable to configure root " "logger") from e

    def configure_formatter(self, config):
        """Configure a formatter from a dictionary."""
        if "()" in config:
            factory = config["()"]  # for use in exception handler
            try:
                result = self.configure_custom(config)
            except TypeError as te:
                if "'format'" not in str(te):
                    raise
                # Name of parameter changed from fmt to format.
                # Retry with old name.
                # This is so that code can be used with older Python versions
                # (e.g. by Django)
                config["fmt"] = config.pop("format")
                config["()"] = factory
                result = self.configure_custom(config)
        else:
            fmt = config.get("format", None)
            dfmt = config.get("datefmt", None)
            style = config.get("style", "%")
            cname = config.get("class", None)

            if not cname:
                c = picologging.Formatter
            else:
                c = _resolve(cname)

            # A TypeError would be raised if "validate" key is passed in with a formatter callable
            # that does not accept "validate" as a parameter
            if (
                "validate" in config
            ):  # if user hasn't mentioned it, the default will be fine
                result = c(fmt, dfmt, style, config["validate"])
            else:
                result = c(fmt, dfmt, style)

        return result

    def configure_filter(self, config):
        """Configure a filter from a dictionary."""
        if "()" in config:
            result = self.configure_custom(config)
        else:
            name = config.get("name", "")
            result = picologging.Filter(name)
        return result

    def add_filters(self, filterer, filters):
        """Add filters to a filterer from a list of names."""
        for f in filters:
            try:
                filterer.addFilter(self.config["filters"][f])
            except Exception as e:
                raise ValueError("Unable to add filter %r" % f) from e

    def configure_handler(self, config):
        """Configure a handler from a dictionary."""
        config_copy = dict(config)  # for restoring in case of error
        formatter = config.pop("formatter", None)
        if formatter:
            try:
                formatter = self.config["formatters"][formatter]
            except Exception as e:
                raise ValueError("Unable to set formatter " "%r" % formatter) from e
        level = config.pop("level", None)
        filters = config.pop("filters", None)
        if "()" in config:
            c = config.pop("()")
            if not callable(c):
                c = self.resolve(c)
            factory = c
        else:
            cname = config.pop("class")
            klass = self.resolve(cname)
            # Special case for handler which refers to another handler
            if issubclass(klass, MemoryHandler) and "target" in config:
                try:
                    th = self.config["handlers"][config["target"]]
                    if not isinstance(th, picologging.Handler):
                        config.update(config_copy)  # restore for deferred cfg
                        raise TypeError("target not configured yet")
                    config["target"] = th
                except Exception as e:
                    raise ValueError(
                        "Unable to set target handler " "%r" % config["target"]
                    ) from e
            # elif (
            #     issubclass(klass, picologging.handlers.SMTPHandler)
            #     and "mailhost" in config
            # ):
            #     config["mailhost"] = self.as_tuple(config["mailhost"])
            # elif (
            #     issubclass(klass, picologging.handlers.SysLogHandler)
            #     and "address" in config
            # ):
            #     config["address"] = self.as_tuple(config["address"])
            factory = klass
        props = config.pop(".", None)
        kwargs = {k: config[k] for k in config if valid_ident(k)}
        try:
            result = factory(**kwargs)
        except TypeError as te:
            if "'stream'" not in str(te):
                raise
            # The argument name changed from strm to stream
            # Retry with old name.
            # This is so that code can be used with older Python versions
            # (e.g. by Django)
            kwargs["strm"] = kwargs.pop("stream")
            result = factory(**kwargs)
        if formatter:
            result.setFormatter(formatter)
        if level is not None:
            result.setLevel(picologging._checkLevel(level))
        if filters:
            self.add_filters(result, filters)
        if props:
            for name, value in props.items():
                setattr(result, name, value)
        return result

    def add_handlers(self, logger, handlers):
        """Add handlers to a logger from a list of names."""
        for h in handlers:
            try:
                logger.addHandler(self.config["handlers"][h])
            except Exception as e:
                raise ValueError("Unable to add handler %r" % h) from e

    def common_logger_config(self, logger, config):
        """
        Perform configuration which is common to root and non-root loggers.
        """
        level = config.get("level", None)
        if level is not None:
            logger.setLevel(picologging._checkLevel(level))

        # Remove any existing handlers
        for h in logger.handlers[:]:
            logger.removeHandler(h)
        handlers = config.get("handlers", None)
        if handlers:
            self.add_handlers(logger, handlers)
        filters = config.get("filters", None)
        if filters:
            self.add_filters(logger, filters)

    def configure_logger(self, name, config):
        """Configure a non-root logger from a dictionary."""
        logger = picologging.getLogger(name)
        self.common_logger_config(logger, config)
        logger.disabled = False
        propagate = config.get("propagate", None)
        if propagate is not None:
            logger.propagate = propagate

    def configure_root(self, config):
        """Configure a root logger from a dictionary."""
        root = picologging.getLogger()
        self.common_logger_config(root, config)


dictConfigClass = DictConfigurator


def dictConfig(config):
    """Configure logging using a dictionary."""
    dictConfigClass(config).configure()



import io

import atheris

with atheris.instrument_imports():
    import sys

    import picologging

    picologging.basicConfig()
    logger = picologging.Logger("fuzz", picologging.DEBUG)
    tmp = io.StringIO()
    logger.addHandler(picologging.StreamHandler(tmp))


def TestOneInput(data):
    fdp = atheris.FuzzedDataProvider(data)
    logger.warning(fdp.ConsumeUnicode(1))


atheris.Setup(sys.argv, TestOneInput)
atheris.Fuzz()



import logging
import sys

import pytest

import picologging

levels = [
    (picologging.DEBUG, "DEBUG"),
    (picologging.INFO, "INFO"),
    (picologging.WARNING, "WARNING"),
    (picologging.ERROR, "ERROR"),
    (picologging.CRITICAL, "CRITICAL"),
    (picologging.NOTSET, "NOTSET"),
]


@pytest.mark.parametrize("level, level_name", levels)
def test_getlevelname(level, level_name):
    assert picologging.getLevelName(level) == level_name
    assert picologging.getLevelName(level_name) == level


def test_value_error_invalid_string_names():
    with pytest.raises(ValueError):
        assert picologging.getLevelName("EXample") == "Level EXample"


junk_level_names = [None, 3.2, (), [], {}, 100]


@pytest.mark.parametrize("level", junk_level_names)
def test_getlevelname_invalid_level(level):
    with pytest.raises((TypeError, ValueError)):
        picologging.getLevelName(level)


def test_root_logger_critical(capsys):
    picologging.root.handlers = []
    picologging.critical("test")

    cap = capsys.readouterr()
    assert cap.out == ""
    assert cap.err == "CRITICAL:root:test\n"


def test_root_logger_fatal(capsys):
    picologging.root.handlers = []
    picologging.fatal("test")

    cap = capsys.readouterr()
    assert cap.out == ""
    assert cap.err == "CRITICAL:root:test\n"


def test_root_logger_error(capsys):
    picologging.root.handlers = []
    picologging.error("test")

    cap = capsys.readouterr()
    assert cap.out == ""
    assert cap.err == "ERROR:root:test\n"


def test_root_logger_exception(capsys):
    picologging.root.handlers = []
    picologging.exception("test", exc_info=Exception("bork bork bork"))

    cap = capsys.readouterr()
    assert cap.out == ""
    assert cap.err == "ERROR:root:test\nException: bork bork bork\n"


def test_root_logger_warning(capsys):
    picologging.root.handlers = []
    picologging.warning("test")

    cap = capsys.readouterr()
    assert cap.out == ""
    assert cap.err == "WARNING:root:test\n"


def test_root_logger_warn(capsys):
    picologging.root.handlers = []
    picologging.warn("test")

    cap = capsys.readouterr()
    assert cap.out == ""
    assert cap.err == "WARNING:root:test\n"


def test_root_logger_info(capsys):
    picologging.root.handlers = []
    picologging.info("test")

    cap = capsys.readouterr()
    assert cap.out == ""
    assert cap.err == ""


def test_root_logger_debug(capsys):
    picologging.root.handlers = []
    picologging.debug("test")

    cap = capsys.readouterr()
    assert cap.out == ""
    assert cap.err == ""


def test_root_logger_log():
    picologging.root.handlers = []
    picologging.log(picologging.DEBUG, "test")


def test_basic_config_with_stream_and_filename_without_handlers():
    picologging.root.handlers = []

    with pytest.raises(ValueError):
        picologging.basicConfig(stream=sys.stderr, filename="log.txt")


def test_basic_config_with_stream_or_filename_with_handlers():
    handler = picologging.StreamHandler(sys.stderr)

    with pytest.raises(ValueError):
        picologging.basicConfig(handlers=[handler], stream=sys.stdout)


def test_basic_config_invalid_style():
    with pytest.raises(ValueError):
        picologging.basicConfig(style="!")


def test_basic_config_with_level():
    picologging.basicConfig(level=picologging.INFO)
    assert picologging.root.level == picologging.INFO


def test_basic_config_invalid_arguments():
    picologging.root.handlers = []
    with pytest.raises(ValueError):
        picologging.basicConfig(invalid_argument="value")


def test_make_log_record():
    log_record = picologging.makeLogRecord({"levelno": picologging.WARNING})

    assert log_record.levelno == picologging.WARNING



import logging
import os
import threading

import pytest

import picologging
from picologging import LogRecord


def test_logrecord_standard():
    record = LogRecord(
        "hello", logging.WARNING, __file__, 123, "bork bork bork", (), None
    )
    assert record.name == "hello"
    assert record.msg == "bork bork bork"
    assert record.levelno == logging.WARNING
    assert record.levelname == "WARNING"
    assert record.pathname == __file__
    assert record.module == "test_logrecord"
    assert record.filename == "test_logrecord.py"
    assert record.args == ()
    assert record.created


def test_logrecord_args():
    record = LogRecord(
        "hello", logging.WARNING, __file__, 123, "bork %s", ("boom"), None
    )
    assert record.name == "hello"
    assert record.msg == "bork %s"
    assert record.args == ("boom")
    assert record.message is None


def test_logrecord_getmessage_with_args():
    record = LogRecord(
        "hello", logging.WARNING, __file__, 123, "bork %s", ("boom"), None
    )
    assert record.message is None
    assert record.getMessage() == "bork boom"
    assert record.message == "bork boom"
    assert record.message == "bork boom"


def test_logrecord_getmessage_no_args():
    record = LogRecord("hello", logging.WARNING, __file__, 123, "bork boom", (), None)
    assert record.message is None
    assert record.getMessage() == "bork boom"
    assert record.message == "bork boom"
    assert record.message == "bork boom"


def test_args_format_mismatch():
    record = LogRecord(
        "hello", logging.WARNING, __file__, 123, "bork boom %s %s", (0,), None
    )
    assert record.message is None
    with pytest.raises(TypeError):
        record.getMessage()


def test_args_len_mismatch():
    record = LogRecord(
        "hello", logging.WARNING, __file__, 123, "bork boom %s", (0, 1, 2), None
    )
    assert record.message is None
    with pytest.raises(TypeError):
        record.getMessage()


def test_no_args():
    record = LogRecord("hello", logging.WARNING, __file__, 123, "bork boom", None, None)
    assert record.message is None
    assert record.getMessage() == "bork boom"
    assert record.message == "bork boom"


def test_no_args_and_format():
    record = LogRecord("hello", logging.WARNING, __file__, 123, "bork %s", None, None)
    assert record.message is None
    assert record.getMessage() == "bork %s"
    assert record.message == "bork %s"


def test_logrecord_single_string_arg():
    record = LogRecord("", picologging.WARNING, "", 12, " %s", "\U000b6fb2", None)
    assert record.args == "\U000b6fb2"
    assert record.getMessage() == " \U000b6fb2"


def test_logrecord_single_empty_string_in_tuple_arg():
    record = LogRecord("", 0, "", 0, " %s", ("",), None)
    assert record.args == ("",)
    assert record.getMessage() == " "


def test_logrecord_single_dict_in_tuple_arg():
    record = LogRecord("", 0, "", 0, "%(key)s", ({"key": "val"},), None)
    assert record.args == {"key": "val"}
    assert record.getMessage() == "val"


def test_logrecord_nested_tuple_arg():
    record = LogRecord("", 0, "", 0, "%d %s", ((10, "bananas"),), None)
    assert record.args == ((10, "bananas"),)
    with pytest.raises(TypeError):
        record.getMessage()


def test_repr():
    record = LogRecord("hello", logging.WARNING, __file__, 123, "bork %s", (0,), None)
    assert repr(record) == f"<LogRecord: hello, 30, {__file__}, 123, 'bork %s'>"


def test_mapping_dict():
    args = {
        "a": "b",
    }
    record = LogRecord(
        "hello", logging.WARNING, __file__, 123, "bork %s", (args,), None
    )
    assert record.args == {"a": "b"}


def test_threading_info():
    record = LogRecord("hello", logging.WARNING, __file__, 123, "bork", (), None)
    assert record.thread == threading.get_ident()
    assert record.threadName is None  # Not supported


def test_process_info():
    record = LogRecord("hello", logging.WARNING, __file__, 123, "bork", (), None)
    assert record.process == os.getpid()
    assert record.processName is None  # Not supported


def test_logrecord_subclass():
    class DerivedLogRecord(LogRecord):
        pass

    record = DerivedLogRecord(
        "hello", logging.WARNING, __file__, 123, "bork boom", (), None
    )

    assert DerivedLogRecord.__base__ is LogRecord
    assert record.message is None
    assert record.getMessage() == "bork boom"
    assert record.message == "bork boom"
    assert record.message == "bork boom"

    handler = picologging.StreamHandler()
    handler.emit(record)



from datetime import datetime
from queue import Queue, SimpleQueue
from socket import socket
from typing import Any, Callable, Pattern

from _typeshed import StrPath

from picologging import FileHandler, Handler, LogRecord

class WatchedFileHandler(FileHandler):
    baseFilename: str  # undocumented
    mode: str  # undocumented
    encoding: str | None  # undocumented
    delay: bool  # undocumented
    errors: str | None  # undocumented
    def __init__(
        self,
        filename: StrPath,
        mode: str = ...,
        encoding: str | None = ...,
        delay: bool = ...,
        errors: str | None = ...,
    ) -> None: ...

class BaseRotatingHandler(FileHandler):
    namer: Callable[[str], str] | None
    rotator: Callable[[str, str], None] | None
    def __init__(
        self,
        filename: StrPath,
        mode: str,
        encoding: str | None = ...,
        delay: bool = ...,
        errors: str | None = ...,
    ) -> None: ...
    def rotation_filename(self, default_name: str) -> str: ...
    def rotate(self, source: str, dest: str) -> None: ...

class RotatingFileHandler(BaseRotatingHandler):
    maxBytes: str  # undocumented
    backupCount: int  # undocumented
    def __init__(
        self,
        filename: StrPath,
        mode: str = ...,
        maxBytes: int = ...,
        backupCount: int = ...,
        encoding: str | None = ...,
        delay: bool = ...,
        errors: str | None = ...,
    ) -> None: ...
    def doRollover(self) -> None: ...
    def shouldRollover(self, record: LogRecord) -> int: ...  # undocumented

class TimedRotatingFileHandler(BaseRotatingHandler):
    when: str  # undocumented
    backupCount: int  # undocumented
    utc: bool  # undocumented
    atTime: datetime.time | None  # undocumented
    interval: int  # undocumented
    suffix: str  # undocumented
    dayOfWeek: int  # undocumented
    rolloverAt: int  # undocumented
    extMatch: Pattern[str]  # undocumented
    def __init__(
        self,
        filename: StrPath,
        when: str = ...,
        interval: int = ...,
        backupCount: int = ...,
        encoding: str | None = ...,
        delay: bool = ...,
        utc: bool = ...,
        atTime: datetime.time | None = ...,
        errors: str | None = ...,
    ) -> None: ...
    def doRollover(self) -> None: ...
    def shouldRollover(self, record: LogRecord) -> int: ...  # undocumented
    def computeRollover(self, currentTime: int) -> int: ...  # undocumented
    def getFilesToDelete(self) -> list[str]: ...  # undocumented

class QueueHandler(Handler):
    queue: SimpleQueue[Any] | Queue[Any]  # undocumented
    def __init__(self, queue: SimpleQueue[Any] | Queue[Any]) -> None: ...

class QueueListener:
    handlers: tuple[Handler, ...]  # undocumented
    respect_handler_level: bool  # undocumented
    queue: SimpleQueue[Any] | Queue[Any]  # undocumented
    def __init__(
        self,
        queue: SimpleQueue[Any] | Queue[Any],
        *handlers: Handler,
        respect_handler_level: bool = ...
    ) -> None: ...
    def dequeue(self, block: bool) -> LogRecord: ...
    def prepare(self, record: LogRecord) -> Any: ...
    def start(self) -> None: ...
    def stop(self) -> None: ...
    def enqueue_sentinel(self) -> None: ...
    def handle(self, record: LogRecord) -> None: ...

class BufferingHandler(Handler):
    capacity: int  # undocumented
    buffer: list[LogRecord]  # undocumented
    def __init__(self, capacity: int) -> None: ...

class MemoryHandler(BufferingHandler):
    flushLevel: int  # undocumented
    target: Handler | None  # undocumented
    flushOnClose: bool  # undocumented
    def __init__(
        self,
        capacity: int,
        flushLevel: int = ...,
        target: Handler | None = ...,
        flushOnClose: bool = ...,
    ) -> None: ...
    def setTarget(self, target: Handler | None) -> None: ...

class SocketHandler(Handler):
    host: str  # undocumented
    port: int | None  # undocumented
    address: tuple[str, int] | str  # undocumented
    sock: socket | None  # undocumented
    closeOnError: bool  # undocumented
    retryTime: float | None  # undocumented
    retryStart: float  # undocumented
    retryFactor: float  # undocumented
    retryMax: float  # undocumented
    def __init__(self, host: str, port: int | None) -> None: ...
    def makeSocket(self, timeout: float = ...) -> socket: ...  # timeout is undocumented
    def makePickle(self, record: LogRecord) -> bytes: ...
    def send(self, s: bytes) -> None: ...
    def createSocket(self) -> None: ...

class DatagramHandler(SocketHandler):
    def makeSocket(self) -> socket: ...



import io
import os
import sys
import warnings
from logging import BufferingFormatter, Filter, StringTemplateStyle, _checkLevel

from ._picologging import Handler  # NOQA
from ._picologging import (
    Filterer,
    FormatStyle,
    Formatter,
    Logger,
    LogRecord,
    StreamHandler,
    getLevelName,
)

__version__ = "0.9.0"

CRITICAL = 50
FATAL = CRITICAL
ERROR = 40
WARNING = 30
WARN = WARNING
INFO = 20
DEBUG = 10
NOTSET = 0

BASIC_FORMAT = "%(levelname)s:%(name)s:%(message)s"


class PercentStyle(FormatStyle):
    def __new__(cls, *args, **kwargs):
        kwargs["style"] = "%"
        return super().__new__(cls, *args, **kwargs)

    def __init__(self, fmt, defaults=None):
        super().__init__(fmt, defaults, style="%")


class StrFormatStyle(FormatStyle):
    def __new__(cls, *args, **kwargs):
        kwargs["style"] = "{"
        return super().__new__(cls, *args, **kwargs)

    def __init__(self, fmt, defaults=None):
        super().__init__(fmt, defaults, style="{")


_STYLES = {
    "%": (PercentStyle, BASIC_FORMAT),
    "{": (StrFormatStyle, "{levelname}:{name}:{message}"),
    "$": (StringTemplateStyle, "${levelname}:${name}:${message}"),
}


class _Placeholder:
    """
    _Placeholder instances are used in the Manager logger hierarchy to take
    the place of nodes for which no loggers have been defined. This class is
    intended for internal use only and not as part of the public API.
    """

    def __init__(self, alogger):
        """
        Initialize with the specified logger being a child of this placeholder.
        """
        self.loggerMap = {alogger: None}

    def append(self, alogger):
        """
        Add the specified logger as a child of this placeholder.
        """
        if alogger not in self.loggerMap:
            self.loggerMap[alogger] = None


class Manager:
    """
    There is [under normal circumstances] just one Manager instance, which
    holds the hierarchy of loggers.
    """

    def __init__(self, rootnode, cls=None):
        """
        Initialize the manager with the root node of the logger hierarchy.
        """
        self.root = rootnode
        self.disable = 0
        self.emittedNoHandlerWarning = False
        self.loggerDict = {}
        if not cls:
            self.cls = Logger
        else:
            self.cls = cls

    @property
    def disable(self):
        return self._disable

    @disable.setter
    def disable(self, value):
        self._disable = _checkLevel(value)

    def getLogger(self, name):
        """
        Get a logger with the specified name (channel name), creating it
        if it doesn't yet exist. This name is a dot-separated hierarchical
        name, such as "a", "a.b", "a.b.c" or similar.
        """
        if name in self.loggerDict:
            rv = self.loggerDict[name]
            if isinstance(rv, _Placeholder):
                ph = rv
                rv = self.cls(name)
                rv.manager = self
                self.loggerDict[name] = rv
                self._fixupChildren(ph, rv)
                self._fixupParents(rv)
        else:
            rv = self.cls(name)
            rv.manager = self
            self.loggerDict[name] = rv
            self._fixupParents(rv)
        return rv

    def _fixupParents(self, alogger):
        """
        Ensure that there are either loggers or placeholders all the way
        from the specified logger to the root of the logger hierarchy.
        """
        name = alogger.name
        i = name.rfind(".")
        logger_parent = None
        while (i > 0) and not logger_parent:
            substr = name[:i]
            if substr not in self.loggerDict:
                self.loggerDict[substr] = _Placeholder(alogger)
            else:
                obj = self.loggerDict[substr]
                if isinstance(obj, Logger):
                    logger_parent = obj
                else:
                    assert isinstance(obj, _Placeholder)
                    obj.append(alogger)
            i = name.rfind(".", 0, i - 1)
        if not logger_parent:
            logger_parent = self.root
        alogger.parent = logger_parent

    def _fixupChildren(self, ph, alogger):
        """
        Ensure that children of the placeholder ph are connected to the
        specified logger.
        """
        name = alogger.name
        namelen = len(name)
        for c in ph.loggerMap.keys():
            # The if means ... if not c.parent.name.startswith(nm)
            if c.parent.name[:namelen] != name:
                alogger.parent = c.parent
                c.parent = alogger

    def setLoggerClass(self, klass):
        self.cls = klass

    def setLogRecordFactory(self, factory):
        raise NotImplementedError("setLoggerClass is not supported in picologging.")


root = Logger(name="root", level=WARNING)
root.manager = Manager(root)


def basicConfig(**kwargs):
    """
    Do basic configuration for the logging system.

    This function does nothing if the root logger already has handlers
    configured, unless the keyword argument *force* is set to ``True``.
    It is a convenience method intended for use by simple scripts
    to do one-shot configuration of the logging package.

    The default behaviour is to create a StreamHandler which writes to
    sys.stderr, set a formatter using the BASIC_FORMAT format string, and
    add the handler to the root logger.

    A number of optional keyword arguments may be specified, which can alter
    the default behaviour.

    filename  Specifies that a FileHandler be created, using the specified
              filename, rather than a StreamHandler.
    filemode  Specifies the mode to open the file, if filename is specified
              (if filemode is unspecified, it defaults to 'a').
    format    Use the specified format string for the handler.
    datefmt   Use the specified date/time format.
    style     If a format string is specified, use this to specify the
              type of format string (possible values '%', '{', '$', for
              %-formatting, :meth:`str.format` and :class:`string.Template`
              - defaults to '%').
    level     Set the root logger level to the specified level.
    stream    Use the specified stream to initialize the StreamHandler. Note
              that this argument is incompatible with 'filename' - if both
              are present, 'stream' is ignored.
    handlers  If specified, this should be an iterable of already created
              handlers, which will be added to the root handler. Any handler
              in the list which does not have a formatter assigned will be
              assigned the formatter created in this function.
    force     If this keyword  is specified as true, any existing handlers
              attached to the root logger are removed and closed, before
              carrying out the configuration as specified by the other
              arguments.
    encoding  If specified together with a filename, this encoding is passed to
              the created FileHandler, causing it to be used when the file is
              opened.
    errors    If specified together with a filename, this value is passed to the
              created FileHandler, causing it to be used when the file is
              opened in text mode. If not specified, the default value is
              `backslashreplace`.

    Note that you could specify a stream created using open(filename, mode)
    rather than passing the filename and mode in. However, it should be
    remembered that StreamHandler does not close its stream (since it may be
    using sys.stdout or sys.stderr), whereas FileHandler closes its stream
    when the handler is closed.

    .. versionchanged:: 3.2
       Added the ``style`` parameter.

    .. versionchanged:: 3.3
       Added the ``handlers`` parameter. A ``ValueError`` is now thrown for
       incompatible arguments (e.g. ``handlers`` specified together with
       ``filename``/``filemode``, or ``filename``/``filemode`` specified
       together with ``stream``, or ``handlers`` specified together with
       ``stream``.

    .. versionchanged:: 3.8
       Added the ``force`` parameter.

    .. versionchanged:: 3.9
       Added the ``encoding`` and ``errors`` parameters.
    """
    # Add thread safety in case someone mistakenly calls
    # basicConfig() from multiple threads
    force = kwargs.pop("force", False)
    encoding = kwargs.pop("encoding", None)
    errors = kwargs.pop("errors", "backslashreplace")
    if force:
        for h in root.handlers[:]:
            root.removeHandler(h)
            h.close()
    if len(root.handlers) == 0:
        handlers = kwargs.pop("handlers", None)
        if handlers is None:
            if "stream" in kwargs and "filename" in kwargs:
                raise ValueError(
                    "'stream' and 'filename' should not be " "specified together"
                )
        else:
            if "stream" in kwargs or "filename" in kwargs:
                raise ValueError(
                    "'stream' or 'filename' should not be "
                    "specified together with 'handlers'"
                )
        if handlers is None:
            filename = kwargs.pop("filename", None)
            mode = kwargs.pop("filemode", "a")
            if filename:
                if "b" in mode:
                    errors = None
                else:
                    encoding = io.text_encoding(encoding)
                h = FileHandler(filename, mode, encoding=encoding, errors=errors)
            else:
                stream = kwargs.pop("stream", sys.stderr)
                h = StreamHandler(stream)
            handlers = [h]
        dfs = kwargs.pop("datefmt", None)
        style = kwargs.pop("style", "%")
        if style not in _STYLES:
            raise ValueError("Style must be one of: %s" % ",".join(_STYLES.keys()))
        fs = kwargs.pop("format", _STYLES[style][1])
        fmt = Formatter(fs, dfs, style)
        for h in handlers:
            if h.formatter is None:
                h.setFormatter(fmt)
            root.addHandler(h)
        level = kwargs.pop("level", None)
        if level is not None:
            root.setLevel(level)
        if kwargs:
            keys = ", ".join(kwargs.keys())
            raise ValueError("Unrecognised argument(s): %s" % keys)


def getLogger(name=None):
    """
    Return a logger with the specified name, creating it if necessary.

    If no name is specified, return the root logger.
    """
    if not name or isinstance(name, str) and name == root.name:
        return root
    return root.manager.getLogger(name)


def critical(msg, *args, **kwargs):
    """
    Log a message with severity 'CRITICAL' on the root logger. If the logger
    has no handlers, call basicConfig() to add a console handler with a
    pre-defined format.
    """
    if len(root.handlers) == 0:
        basicConfig()
    root.critical(msg, *args, **kwargs)


def fatal(msg, *args, **kwargs):
    """
    Don't use this function, use critical() instead.
    """
    critical(msg, *args, **kwargs)


def error(msg, *args, **kwargs):
    """
    Log a message with severity 'ERROR' on the root logger. If the logger has
    no handlers, call basicConfig() to add a console handler with a pre-defined
    format.
    """
    if len(root.handlers) == 0:
        basicConfig()
    root.error(msg, *args, **kwargs)


def exception(msg, *args, exc_info=True, **kwargs):
    """
    Log a message with severity 'ERROR' on the root logger, with exception
    information. If the logger has no handlers, basicConfig() is called to add
    a console handler with a pre-defined format.
    """
    error(msg, *args, exc_info=exc_info, **kwargs)


def warning(msg, *args, **kwargs):
    """
    Log a message with severity 'WARNING' on the root logger. If the logger has
    no handlers, call basicConfig() to add a console handler with a pre-defined
    format.
    """
    if len(root.handlers) == 0:
        basicConfig()
    root.warning(msg, *args, **kwargs)


def warn(msg, *args, **kwargs):
    warnings.warn(
        "The 'warn' function is deprecated, " "use 'warning' instead",
        DeprecationWarning,
        2,
    )
    warning(msg, *args, **kwargs)


def info(msg, *args, **kwargs):
    """
    Log a message with severity 'INFO' on the root logger. If the logger has
    no handlers, call basicConfig() to add a console handler with a pre-defined
    format.
    """
    if len(root.handlers) == 0:
        basicConfig()
    root.info(msg, *args, **kwargs)


def debug(msg, *args, **kwargs):
    """
    Log a message with severity 'DEBUG' on the root logger. If the logger has
    no handlers, call basicConfig() to add a console handler with a pre-defined
    format.
    """
    if len(root.handlers) == 0:
        basicConfig()
    root.debug(msg, *args, **kwargs)


def log(level, msg, *args, **kwargs):
    """
    Log 'msg % args' with the integer severity 'level' on the root logger. If
    the logger has no handlers, call basicConfig() to add a console handler
    with a pre-defined format.
    """
    if len(root.handlers) == 0:
        basicConfig()
    root.log(level, msg, *args, **kwargs)


def disable(level=CRITICAL):
    """
    Disable all logging calls of severity 'level' and below.
    """
    root.manager.disable = level
    root.manager._clear_cache()


class NullHandler(Handler):
    """
    This handler does nothing. It's intended to be used to avoid the
    "No handlers could be found for logger XXX" one-off warning. This is
    important for library code, which may contain code to log events. If a user
    of the library does not configure logging, the one-off warning might be
    produced; to avoid this, the library developer simply needs to instantiate
    a NullHandler and add it to the top-level logger of the library module or
    package.
    """

    def handle(self, record):
        """Stub."""

    def emit(self, record):
        """Stub."""


class FileHandler(StreamHandler):
    """
    A handler class which writes formatted logging records to disk files.
    """

    def __init__(self, filename, mode="a", encoding=None, delay=False):
        """
        Open the specified file and use it as the stream for logging.
        """
        # Issue #27493: add support for Path objects to be passed in
        filename = os.fspath(filename)
        # keep the absolute path, otherwise derived classes which use this
        # may come a cropper when the current directory changes
        self.baseFilename = os.path.abspath(filename)
        self.mode = mode
        self.encoding = encoding
        self.delay = delay
        if delay:
            # We don't open the stream, but we still need to call the
            # Handler constructor to set level, formatter, lock etc.
            Handler.__init__(self)
            self.stream = None
        else:
            StreamHandler.__init__(self, self._open())

    def close(self):
        """
        Closes the stream.
        """
        self.acquire()
        try:
            try:
                if self.stream:
                    try:
                        self.flush()
                    finally:
                        stream = self.stream
                        self.stream = None
                        if hasattr(stream, "close"):
                            stream.close()
            finally:
                # Issue #19523: call unconditionally to
                # prevent a handler leak when delay is set
                StreamHandler.close(self)
        finally:
            self.release()

    def _open(self):
        """
        Open the current base file with the (original) mode and encoding.
        Return the resulting stream.
        """
        return open(self.baseFilename, self.mode, encoding=self.encoding)

    def emit(self, record):
        """
        Emit a record.

        If the stream was not opened because 'delay' was specified in the
        constructor, open it before calling the superclass's emit.
        """
        if self.stream is None:
            self.stream = self._open()
        StreamHandler.emit(self, record)

    def __repr__(self):
        level = getLevelName(self.level)
        return f"<{self.__class__.__name__} {self.baseFilename} ({level})>"


def makeLogRecord(dict):
    """
    Make a LogRecord whose attributes are defined by the specified dictionary,
    This function is useful for converting a logging event received over
    a socket connection (which is sent as a dictionary) into a LogRecord
    instance.
    """

    rv = LogRecord("", NOTSET, "", 0, "", None, None)
    for k, v in dict.items():
        setattr(rv, k, v)
    return rv



import os
import pickle
import queue
import re
import socket
import struct
import threading
import time

import picologging

_MIDNIGHT = 24 * 60 * 60  # number of seconds in a day


class WatchedFileHandler(picologging.FileHandler):
    """
    A handler for logging to a file, which watches the file
    to see if it has changed while in use. This can happen because of
    usage of programs such as newsyslog and logrotate which perform
    log file rotation. This handler, intended for use under Unix,
    watches the file to see if it has changed since the last emit.
    (A file has changed if its device or inode have changed.)
    If it has changed, the old file stream is closed, and the file
    opened to get a new stream.

    This handler is not appropriate for use under Windows, because
    under Windows open files cannot be moved or renamed - logging
    opens the files with exclusive locks - and so there is no need
    for such a handler. Furthermore, ST_INO is not supported under
    Windows; stat always returns zero for this value.
    """

    def __init__(self, filename, mode="a", encoding=None, delay=False):
        picologging.FileHandler.__init__(
            self, filename, mode=mode, encoding=encoding, delay=delay
        )
        self.dev, self.ino = -1, -1
        self._statstream()

    def _statstream(self):
        if self.stream:
            sres = os.fstat(self.stream.fileno())
            self.dev, self.ino = sres.st_dev, sres.st_ino

    def reopenIfNeeded(self):
        """
        Reopen log file if needed.

        Checks if the underlying file has changed, and if it
        has, close the old stream and reopen the file to get the
        current stream.
        """
        try:
            sres = os.stat(self.baseFilename)
        except FileNotFoundError:
            sres = None

        if not sres or sres.st_dev != self.dev or sres.st_ino != self.ino:
            if self.stream is not None:
                self.stream.flush()
                self.stream.close()
                self.stream = None  # See Issue #21742: _open () might fail.
                self.stream = self._open()
                self._statstream()

    def emit(self, record):
        """
        Emit a record.

        If underlying file has changed, reopen the file before emitting the
        record to it.
        """
        self.reopenIfNeeded()
        picologging.FileHandler.emit(self, record)


class BaseRotatingHandler(picologging.FileHandler):
    """
    Base class for handlers that rotate log files at a certain point.
    Not meant to be instantiated directly.  Instead, use RotatingFileHandler
    or TimedRotatingFileHandler.
    """

    namer = None
    rotator = None

    def __init__(self, filename, mode, encoding=None, delay=False):
        """
        Use the specified filename for streamed logging
        """
        picologging.FileHandler.__init__(
            self, filename, mode=mode, encoding=encoding, delay=delay
        )
        self.mode = mode
        self.encoding = encoding

    def shouldRollover(self, record):
        """
        Determine if rollover should occur.
        Should be implemented in inherited classes.
        """

    def doRollover(self, record):
        """
        Do a rollover.
        Should be implemented in inherited classes.
        """

    def emit(self, record):
        """
        Emit a record.
        Output the record to the file, catering for rollover as described
        in doRollover().
        """
        try:
            if self.shouldRollover(record):
                self.doRollover()
            picologging.FileHandler.emit(self, record)
        except Exception:
            self.handleError(record)

    def rotation_filename(self, default_name):
        """
        Modify the filename of a log file when rotating.
        This is provided so that a custom filename can be provided.
        The default implementation calls the 'namer' attribute of the
        handler, if it's callable, passing the default name to
        it. If the attribute isn't callable (the default is None), the name
        is returned unchanged.
        :param default_name: The default name for the log file.
        """
        if not callable(self.namer):
            result = default_name
        else:
            result = self.namer(default_name)
        return result

    def rotate(self, source, dest):
        """
        When rotating, rotate the current log.
        The default implementation calls the 'rotator' attribute of the
        handler, if it's callable, passing the source and dest arguments to
        it. If the attribute isn't callable (the default is None), the source
        is simply renamed to the destination.

        :param source: The source filename. This is normally the base
                       filename, e.g. 'test.log'
        :param dest:   The destination filename. This is normally
                       what the source is rotated to, e.g. 'test.log.1'.
        """
        if not callable(self.rotator):
            # Issue: https://bugs.python.org/issue18940
            # A file may not have been created if delay is True.
            if os.path.exists(source):
                os.rename(source, dest)
        else:
            self.rotator(source, dest)


class RotatingFileHandler(BaseRotatingHandler):
    """
    Handler for logging to a set of files, which switches from one file
    to the next when the current file reaches a certain size.
    """

    def __init__(
        self, filename, mode="a", maxBytes=0, backupCount=0, encoding=None, delay=False
    ):
        """
        Open the specified file and use it as the stream for logging.
        By default, the file grows indefinitely. You can specify particular
        values of maxBytes and backupCount to allow the file to rollover at
        a predetermined size.
        Rollover occurs whenever the current log file is nearly maxBytes in
        length. If backupCount is >= 1, the system will successively create
        new files with the same pathname as the base file, but with extensions
        ".1", ".2" etc. appended to it. For example, with a backupCount of 5
        and a base file name of "app.log", you would get "app.log",
        "app.log.1", "app.log.2", ... through to "app.log.5". The file being
        written to is always "app.log" - when it gets filled up, it is closed
        and renamed to "app.log.1", and if files "app.log.1", "app.log.2" etc.
        exist, then they are renamed to "app.log.2", "app.log.3" etc.
        respectively.
        If maxBytes is zero, rollover never occurs.
        """
        # If rotation/rollover is wanted, it doesn't make sense to use another
        # mode. If for example 'w' were specified, then if there were multiple
        # runs of the calling application, the logs from previous runs would be
        # lost if the 'w' is respected, because the log file would be truncated
        # on each run.
        if maxBytes > 0:
            mode = "a"
        BaseRotatingHandler.__init__(
            self, filename, mode, encoding=encoding, delay=delay
        )
        self.maxBytes = maxBytes
        self.backupCount = backupCount

    def doRollover(self):
        """
        Do a rollover, as described in __init__().
        """
        if self.stream:
            self.stream.close()
            self.stream = None
        if self.backupCount > 0:
            for i in range(self.backupCount - 1, 0, -1):
                sfn = self.rotation_filename("%s.%d" % (self.baseFilename, i))
                dfn = self.rotation_filename("%s.%d" % (self.baseFilename, i + 1))
                if os.path.exists(sfn):
                    if os.path.exists(dfn):
                        os.remove(dfn)
                    os.rename(sfn, dfn)
            dfn = self.rotation_filename(self.baseFilename + ".1")
            if os.path.exists(dfn):
                os.remove(dfn)
            self.rotate(self.baseFilename, dfn)
        if not self.delay:
            self.stream = self._open()

    def shouldRollover(self, record):
        """
        Determine if rollover should occur.
        Basically, see if the supplied record would cause the file to exceed
        the size limit we have.
        """
        # See bpo-45401: Never rollover anything other than regular files
        if os.path.exists(self.baseFilename) and not os.path.isfile(self.baseFilename):
            return False
        if self.stream is None:
            self.stream = self._open()
        if self.maxBytes > 0:
            msg = "%s\n" % self.format(record)
            self.stream.seek(0, 2)  # Due to non-posix-compliant Windows feature
            if self.stream.tell() + len(msg) >= self.maxBytes:
                return True
        return False


class TimedRotatingFileHandler(BaseRotatingHandler):
    """
    Handler for logging to a file, rotating the log file at certain timed
    intervals.
    If backupCount is > 0, when rollover is done, no more than backupCount
    files are kept - the oldest ones are deleted.
    """

    def __init__(
        self,
        filename,
        when="h",
        interval=1,
        backupCount=0,
        encoding=None,
        delay=False,
        utc=False,
        atTime=None,
    ):
        BaseRotatingHandler.__init__(
            self, filename, "a", encoding=encoding, delay=delay
        )
        self.when = when.upper()
        self.backupCount = backupCount
        self.utc = utc
        self.atTime = atTime
        # Calculate the real rollover interval, which is just the number of
        # seconds between rollovers.  Also set the filename suffix used when
        # a rollover occurs.  Current 'when' events supported:
        # S - Seconds
        # M - Minutes
        # H - Hours
        # D - Days
        # midnight - roll over at midnight
        # W{0-6} - roll over on a certain day; 0 - Monday
        #
        # Case of the 'when' specifier is not important; lower or upper case
        # will work.
        if self.when == "S":
            self.interval = 1  # one second
            self.suffix = "%Y-%m-%d_%H-%M-%S"
            self.extMatch = r"^\d{4}-\d{2}-\d{2}_\d{2}-\d{2}-\d{2}(\.\w+)?$"
        elif self.when == "M":
            self.interval = 60  # one minute
            self.suffix = "%Y-%m-%d_%H-%M"
            self.extMatch = r"^\d{4}-\d{2}-\d{2}_\d{2}-\d{2}(\.\w+)?$"
        elif self.when == "H":
            self.interval = 60 * 60  # one hour
            self.suffix = "%Y-%m-%d_%H"
            self.extMatch = r"^\d{4}-\d{2}-\d{2}_\d{2}(\.\w+)?$"
        elif self.when == "D" or self.when == "MIDNIGHT":
            self.interval = 60 * 60 * 24  # one day
            self.suffix = "%Y-%m-%d"
            self.extMatch = r"^\d{4}-\d{2}-\d{2}(\.\w+)?$"
        elif self.when.startswith("W"):
            self.interval = 60 * 60 * 24 * 7  # one week
            if len(self.when) != 2:
                raise ValueError(
                    "You must specify a day for weekly rollover from 0 to 6 (0 is Monday): %s"
                    % self.when
                )
            if self.when[1] < "0" or self.when[1] > "6":
                raise ValueError(
                    "Invalid day specified for weekly rollover: %s" % self.when
                )
            self.dayOfWeek = int(self.when[1])
            self.suffix = "%Y-%m-%d"
            self.extMatch = r"^\d{4}-\d{2}-\d{2}(\.\w+)?$"
        else:
            raise ValueError("Invalid rollover interval specified: %s" % self.when)

        self.extMatch = re.compile(self.extMatch, re.ASCII)
        self.interval = self.interval * interval  # multiply by units requested
        # The following line added because the filename passed in could be a
        # path object (see Issue #27493), but self.baseFilename will be a string
        filename = self.baseFilename
        if os.path.exists(filename):
            t = os.stat(filename).st_mtime
        else:
            t = int(time.time())
        self.rollover_at = self.computeRollover(t)

    def computeRollover(self, current_time):
        """
        Work out the rollover time based on the specified time.
        """
        result = current_time + self.interval
        # If we are rolling over at midnight or weekly, then the interval is already known.
        # What we need to figure out is WHEN the next interval is.  In other words,
        # if you are rolling over at midnight, then your base interval is 1 day,
        # but you want to start that one day clock at midnight, not now.  So, we
        # have to fudge the rollover_at value in order to trigger the first rollover
        # at the right time.  After that, the regular interval will take care of
        # the rest.  Note that this code doesn"t care about leap seconds. :)
        if self.when == "MIDNIGHT" or self.when.startswith("W"):
            # This could be done with less code, but I wanted it to be clear
            if self.utc:
                t = time.gmtime(current_time)
            else:
                t = time.localtime(current_time)
            current_hour = t[3]
            current_minute = t[4]
            current_second = t[5]
            current_day = t[6]
            # r is the number of seconds left between now and the next rotation
            if self.atTime is None:
                rotate_ts = _MIDNIGHT
            else:
                rotate_ts = (
                    self.atTime.hour * 60 + self.atTime.minute
                ) * 60 + self.atTime.second

            r = rotate_ts - ((current_hour * 60 + current_minute) * 60 + current_second)
            if r < 0:
                # Rotate time is before the current time (for example when
                # self.rotateAt is 13:45 and it now 14:15), rotation is
                # tomorrow.
                r += _MIDNIGHT
                current_day = (current_day + 1) % 7
            result = current_time + r
            # If we are rolling over on a certain day, add in the number of days until
            # the next rollover, but offset by 1 since we just calculated the time
            # until the next day starts.  There are three cases:
            # Case 1) The day to rollover is today; in this case, do nothing
            # Case 2) The day to rollover is further in the interval (i.e., today is
            #         day 2 (Wednesday) and rollover is on day 6 (Sunday).  Days to
            #         next rollover is simply 6 - 2 - 1, or 3.
            # Case 3) The day to rollover is behind us in the interval (i.e., today
            #         is day 5 (Saturday) and rollover is on day 3 (Thursday).
            #         Days to rollover is 6 - 5 + 3, or 4.  In this case, it's the
            #         number of days left in the current week (1) plus the number
            #         of days in the next week until the rollover day (3).
            # The calculations described in 2) and 3) above need to have a day added.
            # This is because the above time calculation takes us to midnight on this
            # day, i.e. the start of the next day.
            if self.when.startswith("W"):
                day = current_day  # 0 is Monday
                if day != self.dayOfWeek:
                    if day < self.dayOfWeek:
                        daysToWait = self.dayOfWeek - day
                    else:
                        daysToWait = 6 - day + self.dayOfWeek + 1
                    now_rollover_at = result + (daysToWait * (60 * 60 * 24))
                    if not self.utc:
                        dst_now = t[-1]
                        dstAtRollover = time.localtime(now_rollover_at)[-1]
                        if dst_now != dstAtRollover:
                            if (
                                not dst_now
                            ):  # DST kicks in before next rollover, so we need to deduct an hour
                                addend = -3600
                            else:  # DST bows out before next rollover, so we need to add an hour
                                addend = 3600
                            now_rollover_at += addend
                    result = now_rollover_at
        return result

    def shouldRollover(self, record):
        """
        Determine if rollover should occur.
        record is not used, as we are just comparing times, but it is needed so
        the method signatures are the same
        """
        # See bpo-45401: Never rollover anything other than regular files
        if os.path.exists(self.baseFilename) and not os.path.isfile(self.baseFilename):
            return False
        t = int(time.time())
        if t >= self.rollover_at:
            return True
        return False

    def getFilesToDelete(self):
        """
        Determine the files to delete when rolling over.
        More specific than the earlier method, which just used glob.glob().
        """
        dir_name, base_name = os.path.split(self.baseFilename)
        file_names = os.listdir(dir_name)
        result = []
        # See bpo-44753: Don't use the extension when computing the prefix.
        n, e = os.path.splitext(base_name)
        prefix = n + "."
        plen = len(prefix)
        for file_name in file_names:
            if self.namer is None:
                # Our files will always start with base_name
                if not file_name.startswith(base_name):
                    continue
            else:
                # Our files could be just about anything after custom naming, but
                # likely candidates are of the form
                # foo.log.DATETIME_SUFFIX or foo.DATETIME_SUFFIX.log
                if (
                    not file_name.startswith(base_name)
                    and file_name.endswith(e)
                    and len(file_name) > (plen + 1)
                    and not file_name[plen + 1].isdigit()
                ):
                    continue

            if file_name[:plen] == prefix:
                suffix = file_name[plen:]
                # See bpo-45628: The date/time suffix could be anywhere in the
                # filename
                parts = suffix.split(".")
                for part in parts:
                    if self.extMatch.match(part):
                        result.append(os.path.join(dir_name, file_name))
                        break
        if len(result) < self.backupCount:
            result = []
        else:
            result.sort()
            result = result[: len(result) - self.backupCount]
        return result

    def doRollover(self):
        """
        do a rollover; in this case, a date/time stamp is appended to the filename
        when the rollover happens.  However, you want the file to be named for the
        start of the interval, not the current time.  If there is a backup count,
        then we have to get a list of matching filenames, sort them and remove
        the one with the oldest suffix.
        """
        if self.stream:
            self.stream.close()
            self.stream = None
        # get the time that this sequence started at and make it a TimeTuple
        current_time = int(time.time())
        dst_now = time.localtime(current_time)[-1]
        t = self.rollover_at - self.interval
        if self.utc:
            time_tuple = time.gmtime(t)
        else:
            time_tuple = time.localtime(t)
            dst_then = time_tuple[-1]
            if dst_now != dst_then:
                if dst_now:
                    addend = 3600
                else:
                    addend = -3600
                time_tuple = time.localtime(t + addend)
        dfn = self.rotation_filename(
            self.baseFilename + "." + time.strftime(self.suffix, time_tuple)
        )
        if os.path.exists(dfn):
            os.remove(dfn)
        self.rotate(self.baseFilename, dfn)
        if self.backupCount > 0:
            for s in self.getFilesToDelete():
                os.remove(s)
        if not self.delay:
            self.stream = self._open()
        now_rollover_at = self.computeRollover(current_time)
        while now_rollover_at <= current_time:
            now_rollover_at = now_rollover_at + self.interval
        # If DST changes and midnight or weekly rollover, adjust for this.
        if (self.when == "MIDNIGHT" or self.when.startswith("W")) and not self.utc:
            dstAtRollover = time.localtime(now_rollover_at)[-1]
            if dst_now != dstAtRollover:
                if (
                    not dst_now
                ):  # DST kicks in before next rollover, so we need to deduct an hour
                    addend = -3600
                else:  # DST bows out before next rollover, so we need to add an hour
                    addend = 3600
                now_rollover_at += addend
        self.rollover_at = now_rollover_at


class QueueHandler(picologging.Handler):
    """
    This handler sends events to a queue. Typically, it would be used together
    with a multiprocessing Queue to centralise logging to file in one process
    (in a multi-process application), so as to avoid file write contention
    between processes.

    This code is new in Python 3.2, but this class can be copy pasted into
    user code for use with earlier Python versions.
    """

    def __init__(self, queue):
        """
        Initialise an instance, using the passed queue.
        """
        super().__init__()
        self.queue = queue

    def enqueue(self, record):
        """
        Enqueue a record.

        The base implementation uses put_nowait. You may want to override
        this method if you want to use blocking, timeouts or custom queue
        implementations.
        """
        self.queue.put_nowait(record)

    def prepare(self, record):
        """
        Prepare a record for queuing. The object returned by this method is
        enqueued.

        The base implementation formats the record to merge the message and
        arguments, and removes unpickleable items from the record in-place.
        Specifically, it overwrites the record's `msg` and
        `message` attributes with the merged message (obtained by
        calling the handler's `format` method), and sets the `args`,
        `exc_info` and `exc_text` attributes to None.

        You might want to override this method if you want to convert
        the record to a dict or JSON string, or send a modified copy
        of the record while leaving the original intact.
        """
        # The format operation gets traceback text into record.exc_text
        # (if there's exception data), and also returns the formatted
        # message. We can then use this to replace the original
        # msg + args, as these might be unpickleable. We also zap the
        # exc_info, exc_text and stack_info attributes, as they are no longer
        # needed and, if not None, will typically not be pickleable.
        msg = self.format(record)
        # bpo-35726: make copy of record to avoid affecting other handlers in the chain.
        record = picologging.LogRecord(
            record.name,
            record.levelno,
            record.pathname,
            record.lineno,
            msg,
            None,
            None,
            record.funcName,
            record.stack_info,
        )
        return record

    def emit(self, record: picologging.LogRecord):
        """
        Emit a record.

        Writes the LogRecord to the queue, copying it first.
        """
        try:
            self.enqueue(self.prepare(record))
        except Exception:
            self.handleError(record)


class QueueListener:
    """
    This class implements an internal threaded listener which watches for
    LogRecords being added to a queue, removes them and passes them to a
    list of handlers for processing.
    """

    _sentinel = None

    def __init__(self, queue, *handlers, respect_handler_level=False):
        """
        Initialise an instance with the specified queue and
        handlers.
        """
        self.queue = queue
        self.handlers = handlers
        self._thread = None
        self.respect_handler_level = respect_handler_level

    def dequeue(self, block):
        """
        Dequeue a record and return it, optionally blocking.

        The base implementation uses get. You may want to override this method
        if you want to use timeouts or work with custom queue implementations.
        """
        return self.queue.get(block)

    def start(self):
        """
        Start the listener.

        This starts up a background thread to monitor the queue for
        LogRecords to process.
        """
        self._thread = t = threading.Thread(target=self._monitor)
        t.daemon = True
        t.start()

    def prepare(self, record):
        """
        Prepare a record for handling.

        This method just returns the passed-in record. You may want to
        override this method if you need to do any custom marshalling or
        manipulation of the record before passing it to the handlers.
        """
        return record

    def handle(self, record):
        """
        Handle a record.

        This just loops through the handlers offering them the record
        to handle.
        """
        record = self.prepare(record)
        for handler in self.handlers:
            if not self.respect_handler_level:
                process = True
            else:
                process = record.levelno >= handler.level
            if process:
                handler.handle(record)

    def _monitor(self):
        """
        Monitor the queue for records, and ask the handler
        to deal with them.

        This method runs on a separate, internal thread.
        The thread will terminate if it sees a sentinel object in the queue.
        """
        q = self.queue
        has_task_done = hasattr(q, "task_done")
        while True:
            try:
                record = self.dequeue(True)
                if record is self._sentinel:
                    if has_task_done:
                        q.task_done()
                    break
                self.handle(record)
                if has_task_done:
                    q.task_done()
            except queue.Empty:
                break

    def enqueue_sentinel(self):
        """
        This is used to enqueue the sentinel record.

        The base implementation uses put_nowait. You may want to override this
        method if you want to use timeouts or work with custom queue
        implementations.
        """
        self.queue.put_nowait(self._sentinel)

    def stop(self):
        """
        Stop the listener.

        This asks the thread to terminate, and then waits for it to do so.
        Note that if you don't call this before your application exits, there
        may be some records still left on the queue, which won't be processed.
        """
        self.enqueue_sentinel()
        self._thread.join()
        self._thread = None


class BufferingHandler(picologging.Handler):
    """
    A handler class which buffers logging records in memory. Whenever each
    record is added to the buffer, a check is made to see if the buffer should
    be flushed. If it should, then flush() is expected to do what's needed.
    """

    def __init__(self, capacity):
        """
        Initialize the handler with the buffer size.
        """
        picologging.Handler.__init__(self)
        self.capacity = capacity
        self.buffer = []

    def emit(self, record):
        """
        Emit a record.
        Append the record and call flush() if criteria is met.
        """
        self.buffer.append(record)
        if len(self.buffer) >= self.capacity:
            self.flush()

    def flush(self):
        """
        Override to implement custom flushing behaviour.
        This version just zaps the buffer to empty.
        """
        self.acquire()
        try:
            self.buffer.clear()
        finally:
            self.release()

    def close(self):
        """
        Close the handler.
        This version just flushes and chains to the parent class' close().
        """
        try:
            self.flush()
        finally:
            picologging.Handler.close(self)


class MemoryHandler(BufferingHandler):
    """
    A handler class which buffers logging records in memory, periodically
    flushing them to a target handler. Flushing occurs whenever the buffer
    is full, or when an event of a certain severity or greater is seen.
    """

    def __init__(
        self, capacity, flushLevel=picologging.ERROR, target=None, flushOnClose=True
    ):
        """
        Initialize the handler with the buffer size, the level at which
        flushing should occur and an optional target.
        Note that without a target being set either here or via setTarget(),
        a MemoryHandler is no use to anyone!
        The ``flushOnClose`` argument is ``True`` for backward compatibility
        reasons - the old behaviour is that when the handler is closed, the
        buffer is flushed, even if the flush level hasn't been exceeded nor the
        capacity exceeded. To prevent this, set ``flushOnClose`` to ``False``.
        """
        BufferingHandler.__init__(self, capacity)
        self.flushLevel = flushLevel
        self.target = target
        # See Issue #26559 for why this has been added
        self.flushOnClose = flushOnClose

    def setTarget(self, target):
        """
        Set the target handler for this handler.
        """
        self.acquire()
        try:
            self.target = target
        finally:
            self.release()

    def flush(self):
        """
        For a MemoryHandler, flushing means just sending the buffered
        records to the target, if there is one. Override if you want
        different behaviour.
        The record buffer is also cleared by this operation.
        """
        self.acquire()
        try:
            if self.target:
                for record in self.buffer:
                    self.target.handle(record)
                self.buffer.clear()
        finally:
            self.release()

    def close(self):
        """
        Flush, if appropriately configured, set the target to None and lose the
        buffer.
        """
        try:
            if self.flushOnClose:
                self.flush()
        finally:
            self.acquire()
            try:
                self.target = None
                BufferingHandler.close(self)
            finally:
                self.release()

    def emit(self, record):
        """
        Emit a record.
        Append the record and call flush() if criteria is met.
        """
        self.buffer.append(record)
        if (len(self.buffer) >= self.capacity) or (record.levelno >= self.flushLevel):
            self.flush()


class SocketHandler(picologging.Handler):
    """
    A handler class which writes logging records, in pickle format, to
    a streaming socket. The socket is kept open across logging calls.
    If the peer resets it, an attempt is made to reconnect on the next call.
    The pickle which is sent is that of the LogRecord's attribute dictionary
    (__dict__), so that the receiver does not need to have the logging module
    installed in order to process the logging event.
    To unpickle the record at the receiving end into a LogRecord, use the
    makeLogRecord function.
    """

    def __init__(self, host, port):
        """
        Initializes the handler with a specific host address and port.
        When the attribute *closeOnError* is set to True - if a socket error
        occurs, the socket is silently closed and then reopened on the next
        logging call.
        """
        picologging.Handler.__init__(self)
        self.host = host
        self.port = port
        if port is None:
            self.address = host
        else:
            self.address = (host, port)
        self.sock = None
        self.closeOnError = False
        self.retryTime = None
        # Exponential backoff parameters.
        self.retryStart = 1.0
        self.retryMax = 30.0
        self.retryFactor = 2.0

    def makeSocket(self, timeout=1):
        """
        A factory method which allows subclasses to define the precise
        type of socket they want.
        """
        if self.port is not None:
            result = socket.create_connection(self.address, timeout=timeout)
        else:
            result = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            result.settimeout(timeout)
            try:
                result.connect(self.address)
            except OSError:
                result.close()  # Issue 19182
                raise
        return result

    def createSocket(self):
        """
        Try to create a socket, using an exponential backoff with
        a max retry time.
        """
        now = time.time()
        # Either retryTime is None, in which case this
        # is the first time back after a disconnect, or
        # we've waited long enough.
        if self.retryTime is None:
            attempt = True
        else:
            attempt = now >= self.retryTime
        if attempt:
            try:
                self.sock = self.makeSocket()
                self.retryTime = None  # next time, no delay before trying
            except OSError:
                # Creation failed, so set the retry time and return.
                if self.retryTime is None:
                    self.retryPeriod = self.retryStart
                else:
                    self.retryPeriod = self.retryPeriod * self.retryFactor
                    if self.retryPeriod > self.retryMax:
                        self.retryPeriod = self.retryMax
                self.retryTime = now + self.retryPeriod

    def send(self, s):
        """
        Send a pickled string to the socket.
        This function allows for partial sends which can happen when the
        network is busy.
        """
        if self.sock is None:
            self.createSocket()
        # self.sock can be None either because we haven't reached the retry
        # time yet, or because we have reached the retry time and retried,
        # but are still unable to connect.
        if self.sock:
            try:
                self.sock.sendall(s)
            except OSError:  # pragma: no cover
                self.sock.close()
                self.sock = None  # so we can call createSocket next time

    def makePickle(self, record):
        """
        Pickles the record in binary format with a length prefix, and
        returns it ready for transmission across the socket.
        """
        ei = record.exc_info
        if ei:
            # just to get traceback text into record.exc_text ...
            self.format(record)
        # See issue #14436: If msg or args are objects, they may not be
        # available on the receiving end. So we convert the msg % args
        # to a string, save it as msg and zap the args.
        d = dict(record.__dict__)
        d["msg"] = record.getMessage()
        d["args"] = None
        d["exc_info"] = None
        # Issue #25685: delete 'message' if present: redundant with 'msg'
        d.pop("message", None)
        s = pickle.dumps(d, 1)
        slen = struct.pack(">L", len(s))
        return slen + s

    def handleError(self, record):
        """
        Handle an error during logging.
        An error has occurred during logging. Most likely cause -
        connection lost. Close the socket so that we can retry on the
        next event.
        """
        if self.closeOnError and self.sock:
            self.sock.close()
            self.sock = None  # try to reconnect next time
        else:
            picologging.Handler.handleError(self, record)

    def emit(self, record):
        """
        Emit a record.
        Pickles the record and writes it to the socket in binary format.
        If there is an error with the socket, silently drop the packet.
        If there was a problem with the socket, re-establishes the
        socket.
        """
        try:
            s = self.makePickle(record)
            self.send(s)
        except Exception:
            self.handleError(record)

    def close(self):
        """
        Closes the socket.
        """
        self.acquire()
        try:
            sock = self.sock
            if sock:
                self.sock = None
                sock.close()
            picologging.Handler.close(self)
        finally:
            self.release()


class DatagramHandler(SocketHandler):
    """
    A handler class which writes logging records, in pickle format, to
    a datagram socket. The pickle which is sent is that of the LogRecord's
    attribute dictionary (__dict__), so that the receiver does not need to
    have the logging module installed in order to process the logging event.
    To unpickle the record at the receiving end into a LogRecord, use the
    makeLogRecord function.
    """

    def __init__(self, host, port):
        """
        Initializes the handler with a specific host address and port.
        """
        SocketHandler.__init__(self, host, port)
        self.closeOnError = False

    def makeSocket(self):
        """
        The factory method of SocketHandler is here overridden to create
        a UDP socket (SOCK_DGRAM).
        """
        if self.port is None:
            family = socket.AF_UNIX
        else:
            family = socket.AF_INET
        s = socket.socket(family, socket.SOCK_DGRAM)
        return s

    def send(self, s):
        """
        Send a pickled string to a socket.
        This function no longer allows for partial sends which can happen
        when the network is busy - UDP does not guarantee delivery and
        can deliver packets out of sequence.
        """
        if self.sock is None:
            self.createSocket()
        self.sock.sendto(s, self.address)



import logging
from io import StringIO

import picologging

logging.basicConfig()
picologging.basicConfig()


def record_factory_logging():
    for _ in range(10_000):
        logging.LogRecord(
            "hello", logging.INFO, "/serv/", 123, "bork bork bork", (), None
        )


def record_factory_picologging():
    for _ in range(10_000):
        picologging.LogRecord(
            "hello", logging.INFO, "/serv/", 123, "bork bork bork", (), None
        )


def format_record_logging():
    f = logging.Formatter()
    record = logging.LogRecord(
        "hello", logging.INFO, "/serv/", 123, "bork bork bork", (), None
    )
    for _ in range(10_000):
        f.format(record)


def format_record_picologging():
    f = picologging.Formatter()
    record = picologging.LogRecord(
        "hello", logging.INFO, "/serv/", 123, "bork bork bork", (), None
    )
    for _ in range(10_000):
        f.format(record)


def format_record_with_date_logging():
    f = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    record = logging.LogRecord(
        "hello", logging.INFO, "/serv/", 123, "bork bork bork", (), None
    )
    for _ in range(10_000):
        f.format(record)


def format_record_with_date_picologging():
    f = picologging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    record = picologging.LogRecord(
        "hello", logging.INFO, "/serv/", 123, "bork bork bork", (), None
    )
    for _ in range(10_000):
        f.format(record)


def log_debug_logging(level=logging.DEBUG):
    logger = logging.Logger("test", level)
    tmp = StringIO()

    handler = logging.StreamHandler(tmp)
    handler.setLevel(level)
    formatter = logging.Formatter("%(name)s - %(levelname)s - %(message)s")
    handler.setFormatter(formatter)
    logger.handlers.append(handler)

    for _ in range(10_000):
        logger.debug("There has been a logging issue")


def log_debug_logging_with_args(level=logging.DEBUG):
    logger = logging.Logger("test", level)
    tmp = StringIO()

    handler = logging.StreamHandler(tmp)
    handler.setLevel(level)
    formatter = logging.Formatter("%(name)s - %(levelname)s - %(message)s")
    handler.setFormatter(formatter)
    logger.handlers.append(handler)

    for _ in range(10_000):
        logger.debug("There has been a logging issue %s %s %s", 1, 2, 3)


def log_debug_picologging(level=logging.DEBUG):
    logger = picologging.Logger("test", level)
    tmp = StringIO()

    handler = picologging.StreamHandler(tmp)
    handler.setLevel(level)
    formatter = picologging.Formatter("%(name)s - %(levelname)s - %(message)s")
    handler.setFormatter(formatter)
    logger.handlers.append(handler)

    for _ in range(10_000):
        logger.debug("There has been a picologging issue")


def log_debug_picologging_with_args(level=logging.DEBUG):
    logger = picologging.Logger("test", level)
    tmp = StringIO()

    handler = picologging.StreamHandler(tmp)
    handler.setLevel(level)
    formatter = picologging.Formatter("%(name)s - %(levelname)s - %(message)s")
    handler.setFormatter(formatter)
    logger.handlers.append(handler)

    for _ in range(10_000):
        logger.debug("There has been a picologging issue %s %s %s", 1, 2, 3)


def log_debug_outofscope_logging():
    log_debug_logging(logging.INFO)


def log_debug_outofscope_picologging():
    log_debug_picologging(logging.INFO)


def log_debug_outofscope_logging_with_args():
    log_debug_logging_with_args(logging.INFO)


def log_debug_outofscope_picologging_with_args():
    log_debug_picologging_with_args(logging.INFO)


__benchmarks__ = [
    (record_factory_logging, record_factory_picologging, "LogRecord()"),
    (format_record_logging, format_record_picologging, "Formatter().format()"),
    (
        format_record_with_date_logging,
        format_record_with_date_picologging,
        "Formatter().format() with date",
    ),
    (log_debug_logging, log_debug_picologging, "Logger(level=DEBUG).debug()"),
    (
        log_debug_logging_with_args,
        log_debug_picologging_with_args,
        "Logger(level=DEBUG).debug() with args",
    ),
    (
        log_debug_outofscope_logging,
        log_debug_outofscope_picologging,
        "Logger(level=INFO).debug()",
    ),
    (
        log_debug_outofscope_logging_with_args,
        log_debug_outofscope_picologging_with_args,
        "Logger(level=INFO).debug() with args",
    ),
]



from collections.abc import Callable, Iterable, Mapping
from io import TextIOWrapper
from multiprocessing import Manager
from string import Template
from types import TracebackType
from typing import Any, Generic, Optional, Pattern, TextIO, TypeVar, Union, overload

from _typeshed import StrPath, SupportsWrite
from typing_extensions import Literal, TypeAlias

CRITICAL: int
FATAL: int
ERROR: int
WARNING: int
WARN: int
INFO: int
DEBUG: int
NOTSET: int

_SysExcInfoType: TypeAlias = Union[
    tuple[type[BaseException], BaseException, TracebackType | None],
    tuple[None, None, None],
]
_ExcInfoType: TypeAlias = None | bool | _SysExcInfoType | BaseException
_ArgsType: TypeAlias = tuple[object, ...] | Mapping[str, object]
_Level: TypeAlias = int  # Not | str like it is in logging
_FormatStyle: TypeAlias = Literal["%", "{", "$"]

class LogRecord:
    """
    A LogRecord instance represents an event being logged.

    LogRecord instances are created every time something is logged. They
    contain all the information pertinent to the event being logged. The
    main information passed in is in msg and args, which are combined
    using str(msg) % args to create the message field of the record. The
    record also includes information such as when the record was created,
    the source line where the logging call was made, and any exception
    information to be logged.
    """

    name: str
    levelno: int
    levelname: str
    msg: str
    args: Iterable[Any]
    pathname: str
    filename: str
    module: str
    lineno: int
    funcName: str
    created: float
    msecs: int
    relativeCreated: float
    thread: Optional[int]
    threadName: Optional[str]
    processName: Optional[str]
    process: Optional[int]
    exc_info: Any
    exc_text: Optional[str]
    stack_info: Optional[Any]
    message: str
    asctime: str
    def __init__(
        self,
        name: str,
        level: int,
        pathname: str,
        lineno: int,
        msg: str,
        args: Iterable[Any],
        exc_info: Any,
        func=None,
        sinfo=None,
        **kwargs,
    ): ...
    def getMessage(self) -> str:
        """
        Return the message for this LogRecord.

        Return the message for this LogRecord after merging any user-supplied
        arguments with the message.
        """
        ...

class Formatter:
    datefmt: str
    def __init__(
        self,
        fmt: str | None = ...,
        datefmt: str | None = ...,
        style: _FormatStyle = ...,
        validate: bool = ...,
        *,
        defaults: Mapping[str, Any] | None = ...,
    ) -> None: ...
    def format(self, record: LogRecord) -> str: ...
    def formatMessage(self, record: LogRecord) -> str: ...  # undocumented
    def formatStack(self, stack_info: str) -> str: ...
    def formatException(self, ei: _SysExcInfoType) -> str: ...
    def usesTime(self) -> bool: ...  # undocumented

_FilterType: TypeAlias = Filter | Callable[[LogRecord], int]

class Filterer:
    filters: list[Filter]
    def __init__(self) -> None: ...
    def addFilter(self, filter: _FilterType) -> None: ...
    def removeFilter(self, filter: _FilterType) -> None: ...
    def filter(self, record: LogRecord) -> bool: ...

class Handler(Filterer):
    level: int  # undocumented
    formatter: Formatter | None  # undocumented
    name: str | None  # undocumented
    def __init__(
        self, name: Optional[str] = None, level: Optional[int] = NOTSET
    ) -> None: ...
    def acquire(self) -> None: ...
    def release(self) -> None: ...
    def setLevel(self, level: int) -> None: ...
    def setFormatter(self, fmt: Formatter | None) -> None: ...
    def filter(self, record: LogRecord) -> bool: ...
    def handle(self, record: LogRecord) -> bool: ...
    def format(self, record: LogRecord) -> str: ...
    def emit(self, record: LogRecord) -> None: ...
    def flush(self) -> None: ...
    def close(self) -> None: ...
    def handleError(self, record: LogRecord) -> None: ...
    def get_name(self) -> str: ...
    def set_name(self, name: str) -> None: ...
    def createLock(self) -> None: ...

class Logger(Filterer):
    propagate: bool
    name: str
    level: int
    parent: Logger
    handlers: list[Handler]
    disabled: bool
    manager: Optional[Manager]
    def __init__(self, name: str, level: _Level = ...) -> None: ...
    def setLevel(self, level: _Level) -> None: ...
    def getEffectiveLevel(self) -> int: ...
    def debug(
        self,
        msg: object,
        *args: object,
        exc_info: _ExcInfoType = ...,
        stack_info: bool = ...,
        stacklevel: int = ...,
        extra: Mapping[str, object] | None = ...,
    ) -> None: ...
    def info(
        self,
        msg: object,
        *args: object,
        exc_info: _ExcInfoType = ...,
        stack_info: bool = ...,
        stacklevel: int = ...,
        extra: Mapping[str, object] | None = ...,
    ) -> None: ...
    def warning(
        self,
        msg: object,
        *args: object,
        exc_info: _ExcInfoType = ...,
        stack_info: bool = ...,
        stacklevel: int = ...,
        extra: Mapping[str, object] | None = ...,
    ) -> None: ...
    def warn(
        self,
        msg: object,
        *args: object,
        exc_info: _ExcInfoType = ...,
        stack_info: bool = ...,
        stacklevel: int = ...,
        extra: Mapping[str, object] | None = ...,
    ) -> None: ...
    def error(
        self,
        msg: object,
        *args: object,
        exc_info: _ExcInfoType = ...,
        stack_info: bool = ...,
        stacklevel: int = ...,
        extra: Mapping[str, object] | None = ...,
    ) -> None: ...
    def exception(
        self,
        msg: object,
        *args: object,
        exc_info: _ExcInfoType = ...,
        stack_info: bool = ...,
        stacklevel: int = ...,
        extra: Mapping[str, object] | None = ...,
    ) -> None: ...
    def critical(
        self,
        msg: object,
        *args: object,
        exc_info: _ExcInfoType = ...,
        stack_info: bool = ...,
        stacklevel: int = ...,
        extra: Mapping[str, object] | None = ...,
    ) -> None: ...
    def log(
        self,
        level: int,
        msg: object,
        *args: object,
        exc_info: _ExcInfoType = ...,
        stack_info: bool = ...,
        stacklevel: int = ...,
        extra: Mapping[str, object] | None = ...,
    ) -> None: ...
    fatal = critical
    def filter(self, record: LogRecord) -> bool: ...
    def addHandler(self, hdlr: Handler) -> None: ...
    def removeHandler(self, hdlr: Handler) -> None: ...
    def handle(self, record: LogRecord) -> None: ...

class Filter:
    name: str  # undocumented
    nlen: int  # undocumented
    def __init__(self, name: str = ...) -> None: ...
    def filter(self, record: LogRecord) -> bool: ...

def getLogger(name: str | None = ...) -> Logger: ...
def debug(
    msg: object,
    *args: object,
    exc_info: _ExcInfoType = ...,
    stack_info: bool = ...,
    stacklevel: int = ...,
    extra: Mapping[str, object] | None = ...,
) -> None: ...
def info(
    msg: object,
    *args: object,
    exc_info: _ExcInfoType = ...,
    stack_info: bool = ...,
    stacklevel: int = ...,
    extra: Mapping[str, object] | None = ...,
) -> None: ...
def warning(
    msg: object,
    *args: object,
    exc_info: _ExcInfoType = ...,
    stack_info: bool = ...,
    stacklevel: int = ...,
    extra: Mapping[str, object] | None = ...,
) -> None: ...
def warn(
    msg: object,
    *args: object,
    exc_info: _ExcInfoType = ...,
    stack_info: bool = ...,
    stacklevel: int = ...,
    extra: Mapping[str, object] | None = ...,
) -> None: ...
def error(
    msg: object,
    *args: object,
    exc_info: _ExcInfoType = ...,
    stack_info: bool = ...,
    stacklevel: int = ...,
    extra: Mapping[str, object] | None = ...,
) -> None: ...
def critical(
    msg: object,
    *args: object,
    exc_info: _ExcInfoType = ...,
    stack_info: bool = ...,
    stacklevel: int = ...,
    extra: Mapping[str, object] | None = ...,
) -> None: ...
def exception(
    msg: object,
    *args: object,
    exc_info: _ExcInfoType = ...,
    stack_info: bool = ...,
    stacklevel: int = ...,
    extra: Mapping[str, object] | None = ...,
) -> None: ...
def log(
    level: int,
    msg: object,
    *args: object,
    exc_info: _ExcInfoType = ...,
    stack_info: bool = ...,
    stacklevel: int = ...,
    extra: Mapping[str, object] | None = ...,
) -> None: ...
def basicConfig(
    *,
    filename: StrPath | None = ...,
    filemode: str = ...,
    format: str = ...,
    datefmt: str | None = ...,
    style: _FormatStyle = ...,
    level: _Level | None = ...,
    stream: SupportsWrite[str] | None = ...,
    handlers: Iterable[Handler] | None = ...,
    force: bool | None = ...,
    encoding: str | None = ...,
    errors: str | None = ...,
) -> None: ...

lastResort: StreamHandler[Any] | None

_StreamT = TypeVar("_StreamT", bound=SupportsWrite[str])

class StreamHandler(Handler, Generic[_StreamT]):
    stream: _StreamT  # undocumented
    @overload
    def __init__(self: StreamHandler[TextIO], stream: None = ...) -> None: ...
    @overload
    def __init__(self: StreamHandler[_StreamT], stream: _StreamT) -> None: ...
    def setStream(self, stream: _StreamT) -> _StreamT | None: ...

class FileHandler(StreamHandler[TextIOWrapper]):
    baseFilename: str  # undocumented
    mode: str  # undocumented
    encoding: str | None  # undocumented
    delay: bool  # undocumented
    errors: str | None  # undocumented
    def __init__(
        self,
        filename: StrPath,
        mode: str = ...,
        encoding: str | None = ...,
        delay: bool = ...,
        errors: str | None = ...,
    ) -> None: ...

class NullHandler(Handler): ...

root: Logger

class PercentStyle:
    def __init__(
        self, fmt: str, *, defaults: Mapping[str, Any] | None = ...
    ) -> None: ...
    def usesTime(self) -> bool: ...
    def validate(self) -> None: ...
    def format(self, record: Any) -> str: ...

class StrFormatStyle(PercentStyle):  # undocumented
    fmt_spec: Pattern[str]
    field_spec: Pattern[str]

class StringTemplateStyle(PercentStyle):  # undocumented
    _tpl: Template

_STYLES: dict[str, tuple[PercentStyle, str]]

BASIC_FORMAT: str

def getLevelName(level: _Level) -> Any: ...
def makeLogRecord(dict: Mapping[str, object]) -> LogRecord: ...



import pytest

import picologging


def test_basic_handler():
    handler = picologging.Handler()
    record = picologging.LogRecord(
        "test", picologging.INFO, "test", 1, "test", (), None, None, None
    )
    with pytest.raises(NotImplementedError):
        handler.handle(record)
    with pytest.raises(NotImplementedError):
        handler.emit(None)


def test_custom_handler():
    class CustomHandler(picologging.Handler):
        def __init__(self):
            super().__init__()
            self.records = []

        def emit(self, record):
            self.records.append(record)

    handler = CustomHandler()
    record = picologging.LogRecord(
        "test", picologging.INFO, __file__, 1, "test", (), None, None, None
    )
    formatter = picologging.Formatter("%(message)s")
    handler.setFormatter(formatter)
    handler.handle(record)
    assert len(handler.records) == 1
    assert handler.records[0] == record


def test_delete_handler():
    handler = picologging.Handler()
    del handler


def test_add_acquire_release():
    handler = picologging.Handler()
    handler.acquire()
    assert handler.release() is None


def test_init_with_name():
    handler = picologging.Handler(name="test")
    assert handler.name == "test"


def test_init_with_level():
    handler = picologging.Handler(level=picologging.DEBUG)
    assert handler.level == picologging.DEBUG


def test_get_set_name():
    handler = picologging.Handler(name="test")
    assert handler.get_name() == "test"
    handler.set_name("foo")
    assert handler.name == "foo"
    assert handler.get_name() == "foo"


def test_flush():
    handler = picologging.Handler()
    assert not handler.flush()


def test_close():
    handler = picologging.Handler()
    assert not handler.close()


def test_createLock():
    handler = picologging.Handler()
    assert not handler.createLock()


def test_filtered_out():
    def filter_out(f):
        return False

    class CustomHandler(picologging.Handler):
        def __init__(self):
            super().__init__()
            self.records = []

        def emit(self, record):
            raise Exception("This should not be called")

    handler = CustomHandler()
    handler.addFilter(filter_out)
    record = picologging.LogRecord(
        "test", picologging.INFO, __file__, 1, "test", (), None, None, None
    )
    assert not handler.handle(record)


def test_set_level_nonint():
    handler = picologging.Handler()
    with pytest.raises(TypeError):
        handler.setLevel("potato")


def test_custom_formatter():
    class CustomFormatter:
        def format(self, record):
            return "foo"

    handler = picologging.Handler()
    handler.setFormatter(CustomFormatter())
    record = picologging.LogRecord(
        "test", picologging.INFO, __file__, 1, "test", (), None, None, None
    )
    assert handler.format(record) == "foo"


def test_handle_error():
    handler = picologging.Handler()
    record = picologging.LogRecord(
        "test", picologging.INFO, __file__, 1, "test", (), None, None, None
    )
    assert not handler.handleError(record)


def test_handler_repr():
    handler = picologging.Handler()
    assert repr(handler) == "<Handler (NOTSET)>"

    handler = picologging.Handler(level=picologging.WARNING)
    assert repr(handler) == "<Handler (WARNING)>"



import io
import queue

import picologging
from picologging.handlers import QueueHandler, QueueListener


def test_queue_handler_dispatch():
    logger = picologging.Logger("test", picologging.DEBUG)
    q = queue.Queue()
    handler = QueueHandler(q)
    logger.addHandler(handler)
    logger.debug("test")
    record = q.get(block=False)
    assert record
    assert record.levelno == picologging.DEBUG
    assert record.name == "test"
    assert record.msg == "test"
    assert record.args is None
    assert record.exc_info is None


def test_queue_listener():
    logger = picologging.Logger("test", picologging.DEBUG)
    stream = io.StringIO()
    stream_handler = picologging.StreamHandler(stream)
    q = queue.Queue()
    listener = QueueListener(q, stream_handler)
    listener.start()
    handler = QueueHandler(q)
    logger.addHandler(handler)
    logger.debug("test")

    listener.stop()
    assert stream.getvalue() == "test\n"


def test_queue_handler_handle_exception():
    logger = picologging.Logger("test", picologging.DEBUG)
    q = queue.Queue(maxsize=1)
    handler = QueueHandler(q)
    logger.addHandler(handler)

    handler.queue = None
    logger.debug("test")


def test_queue_handler_format():
    logger = picologging.getLogger("picologging_test")
    logger.setLevel(picologging.INFO)
    stream = io.StringIO()
    stream_handler = picologging.StreamHandler(stream)
    q = queue.Queue()
    listener = QueueListener(q, stream_handler)
    listener.start()
    handler = QueueHandler(q)
    handler.setLevel(picologging.DEBUG)
    handler.setFormatter(
        picologging.Formatter("%(levelname)s - %(name)s - %(message)s")
    )
    logger.addHandler(handler)
    logger.info("Testing now!")

    listener.stop()
    assert stream.getvalue() == "INFO - picologging_test - Testing now!\n"



import os
import pickle
import socket
import struct
import tempfile
import threading
from socketserver import (
    DatagramRequestHandler,
    StreamRequestHandler,
    ThreadingTCPServer,
    ThreadingUDPServer,
)

import pytest

import picologging
from picologging.handlers import DatagramHandler, SocketHandler


class ControlMixin:
    def __init__(self, handler, poll_interval):
        self._thread = None
        self._handler = handler
        self.poll_interval = poll_interval
        self.ready = threading.Event()

    def start(self):
        self._thread = threading.Thread(
            target=self.serve_forever, args=(self.poll_interval,)
        )
        self._thread.daemon = True
        self._thread.start()

    def serve_forever(self, poll_interval):
        self.ready.set()
        super().serve_forever(poll_interval)

    def stop(self):
        self.shutdown()
        if self._thread is not None:
            self._thread.join()
            self._thread = None
        self.server_close()
        self.ready.clear()


class TCPServer(ControlMixin, ThreadingTCPServer):
    allow_reuse_address = True

    def __init__(self, addr, poll_interval=0.5, bind_and_activate=True):
        class DelegatingTCPRequestHandler(StreamRequestHandler):
            def handle(self):
                self.server._handler(self)

        ThreadingTCPServer.__init__(
            self, addr, DelegatingTCPRequestHandler, bind_and_activate
        )
        ControlMixin.__init__(self, self.handle_socket, poll_interval)
        self.log_output = ""
        self.handled = threading.Semaphore(0)

    def server_bind(self):
        super().server_bind()
        self.port = self.socket.getsockname()[1]

    def handle_socket(self, request):
        conn = request.connection
        while True:
            chunk = conn.recv(4)
            if len(chunk) < 4:
                break
            slen = struct.unpack(">L", chunk)[0]
            chunk = conn.recv(slen)
            while len(chunk) < slen:
                chunk = chunk + conn.recv(slen - len(chunk))
            obj = pickle.loads(chunk)
            record = picologging.makeLogRecord(obj)
            self.log_output += record.msg + "\n"
            self.handled.release()


class UDPServer(ControlMixin, ThreadingUDPServer):
    def __init__(self, addr, poll_interval=0.5, bind_and_activate=True):
        class DelegatingUDPRequestHandler(DatagramRequestHandler):
            def handle(self):
                self.server._handler(self)

            def finish(self):
                data = self.wfile.getvalue()
                if data:
                    try:
                        super().finish()
                    except OSError:
                        if not self.server._closed:
                            raise

        ThreadingUDPServer.__init__(
            self, addr, DelegatingUDPRequestHandler, bind_and_activate
        )
        ControlMixin.__init__(self, self.handle_socket, poll_interval)
        self._closed = False
        self.log_output = ""
        self.handled = threading.Semaphore(0)

    def handle_socket(self, request):
        slen = struct.pack(">L", 0)  # length of prefix
        packet = request.packet[len(slen) :]
        obj = pickle.loads(packet)
        record = picologging.makeLogRecord(obj)
        self.log_output += record.msg + "\n"
        self.handled.release()

    def server_bind(self):
        super().server_bind()
        self.port = self.socket.getsockname()[1]

    def server_close(self):
        super().server_close()
        self._closed = True


if hasattr(socket, "AF_UNIX"):

    class UnixStreamServer(TCPServer):
        address_family = socket.AF_UNIX

    class UnixDatagramServer(UDPServer):
        address_family = socket.AF_UNIX


def test_sockethandler():
    server = TCPServer(("localhost", 0), 0.01)
    server.start()
    server.ready.wait()

    handler = SocketHandler("localhost", server.port)
    logger = picologging.getLogger("test")
    logger.setLevel(picologging.DEBUG)
    logger.addHandler(handler)

    logger.error("test")
    server.handled.acquire()
    logger.debug("test")
    server.handled.acquire()

    assert server.log_output == "test\ntest\n"

    handler.close()
    server.stop()


@pytest.mark.skipif(not hasattr(socket, "AF_UNIX"), reason="Unix sockets required")
def test_unix_sockethandler():
    address = tempfile.NamedTemporaryFile(prefix="picologging_", suffix=".sock").name
    server = UnixStreamServer(address, 0.01)
    server.start()
    server.ready.wait()

    handler = SocketHandler(address, None)
    logger = picologging.getLogger("test")
    logger.setLevel(picologging.DEBUG)
    logger.addHandler(handler)

    logger.error("test")
    server.handled.acquire()
    logger.debug("test")
    server.handled.acquire()

    assert server.log_output == "test\ntest\n"

    handler.close()
    server.stop()
    os.remove(address)


@pytest.mark.skipif(not hasattr(socket, "AF_UNIX"), reason="Unix sockets required")
def test_unix_sockethandler_connect_exception(monkeypatch):
    def mock_os_error(*args):
        raise OSError()

    monkeypatch.setattr(socket.socket, "connect", mock_os_error)

    address = tempfile.NamedTemporaryFile(prefix="picologging_", suffix=".sock").name
    handler = SocketHandler(address, None)
    handler.retryMax = 1
    logger = picologging.getLogger("test")
    logger.setLevel(picologging.DEBUG)
    logger.addHandler(handler)

    logger.debug("test")

    handler.retryTime -= 5
    logger.debug("test")

    assert handler.retryTime is not None
    assert handler.sock is None

    handler.close()


@pytest.mark.skipif(not hasattr(socket, "AF_UNIX"), reason="Unix sockets required")
def test_unix_sockethandler_emit_exception(monkeypatch):
    def mock_exception(*args):
        raise Exception()

    monkeypatch.setattr(socket.socket, "sendall", mock_exception)

    address = tempfile.NamedTemporaryFile(prefix="picologging_", suffix=".sock").name
    server = UnixStreamServer(address, 0.01)
    server.start()
    server.ready.wait()

    handler = SocketHandler(address, None)
    logger = picologging.getLogger("test")
    logger.setLevel(picologging.DEBUG)
    logger.addHandler(handler)

    logger.debug("test")

    try:
        1 / 0
    except ZeroDivisionError:
        logger.exception("error")

    handler.closeOnError = True
    logger.debug("test")
    assert handler.sock is None

    handler.close()
    server.stop()
    os.remove(address)


def test_datagramhandler():
    server = UDPServer(("localhost", 0), 0.01)
    server.start()
    server.ready.wait()

    handler = DatagramHandler("localhost", server.port)
    logger = picologging.getLogger("test")
    logger.setLevel(picologging.DEBUG)
    logger.addHandler(handler)

    logger.error("test")
    server.handled.acquire()
    logger.debug("test")
    server.handled.acquire()

    assert server.log_output == "test\ntest\n"

    handler.close()
    server.stop()


@pytest.mark.skipif(not hasattr(socket, "AF_UNIX"), reason="Unix sockets required")
def test_unix_datagramhandler():
    address = tempfile.NamedTemporaryFile(prefix="picologging_", suffix=".sock").name
    server = UnixDatagramServer(address, 0.01)
    server.start()
    server.ready.wait()

    handler = DatagramHandler(address, None)
    logger = picologging.getLogger("test")
    logger.setLevel(picologging.DEBUG)
    logger.addHandler(handler)

    logger.error("test")
    server.handled.acquire()
    logger.debug("test")
    server.handled.acquire()

    assert server.log_output == "test\ntest\n"

    handler.close()
    server.stop()
    os.remove(address)



from io import StringIO

import picologging


def run_profile(level=picologging.DEBUG):
    logger = picologging.Logger("test", level)
    tmp = StringIO()

    handler = picologging.StreamHandler(tmp)
    handler.setLevel(level)
    formatter = picologging.Formatter("%(name)s - %(levelname)s - %(message)s")
    handler.setFormatter(formatter)
    logger.handlers.append(handler)

    for _ in range(100_000):
        logger.debug("There has been a picologging issue %s %s %s", 1, 2, 3)

    assert tmp.getvalue() != ""


run_profile()



import picologging
from picologging.handlers import BufferingHandler, MemoryHandler


def test_buffering_handler():
    logger = picologging.Logger("test", picologging.DEBUG)
    handler = BufferingHandler(capacity=1)
    logger.addHandler(handler)

    logger.debug("test")
    handler.close()

    assert handler.buffer == []


def test_memory_handler(tmp_path):
    log_file = tmp_path / "log.txt"
    target = picologging.FileHandler(log_file)
    logger = picologging.Logger("test", picologging.DEBUG)
    handler = MemoryHandler(capacity=1, target=target)
    logger.addHandler(handler)

    logger.debug("test")
    handler.close()

    with open(log_file) as f:
        assert f.read() == "test\n"
    assert handler.buffer == []


def test_memory_handler_set_target(tmp_path):
    log_file = tmp_path / "log.txt"
    target = picologging.FileHandler(log_file)
    logger = picologging.Logger("test", picologging.DEBUG)
    handler = MemoryHandler(capacity=1)
    handler.setTarget(target)
    logger.addHandler(handler)

    logger.debug("test")
    handler.close()

    with open(log_file) as f:
        assert f.read() == "test\n"
    assert handler.buffer == []



import io
import sys

import pytest

import picologging


def test_stream_handler():
    stream = io.StringIO()
    handler = picologging.StreamHandler(stream)
    record = picologging.LogRecord(
        "test", picologging.INFO, __file__, 1, "test", (), None, None, None
    )
    formatter = picologging.Formatter("%(message)s")
    handler.setFormatter(formatter)
    assert handler.formatter == formatter
    handler.handle(record)
    assert stream.getvalue() == "test\n"


def test_stream_handler_defaults_to_stderr():
    handler = picologging.StreamHandler()
    assert handler.stream == sys.stderr

    handler = picologging.StreamHandler(None)
    assert handler.stream == sys.stderr


def test_stream_handler_bad_init_args():
    with pytest.raises(TypeError):
        picologging.StreamHandler(1, 2, 3, 4)

    with pytest.raises(TypeError):
        picologging.StreamHandler(dog=1)


def test_stream_handler_invalid_stream_type():
    handler = picologging.StreamHandler("potato")
    record = picologging.LogRecord(
        "test", picologging.INFO, __file__, 1, "test", (), None, None, None
    )
    with pytest.raises(AttributeError):
        handler.handle(record)


def test_non_flushable_stream():
    class TestStream:
        def write(self, data):
            pass

    handler = picologging.StreamHandler(TestStream())
    assert not handler.flush()


def test_emit_no_args():
    handler = picologging.StreamHandler()
    with pytest.raises(ValueError):
        handler.emit()


def test_emit_invalid_args_type():
    handler = picologging.StreamHandler()
    with pytest.raises(TypeError):
        handler.emit(1234)


def test_stream_write_raises_error():
    class TestStream:
        def write(self, data):
            raise Exception("blerg")

    handler = picologging.StreamHandler(TestStream())
    with pytest.raises(Exception):
        handler.emit("foo 123")


def test_set_stream():
    class TestStream:
        def write(self, data):
            pass

        def flush(self):
            pass

    a = TestStream()
    handler = picologging.StreamHandler(a)
    assert handler.stream is a
    b = TestStream()
    handler.setStream(b)
    assert handler.stream is b


def test_set_stream_return_value():
    h = picologging.StreamHandler()
    stream = io.StringIO()
    old = h.setStream(stream)
    assert old is sys.stderr
    actual = h.setStream(old)
    assert actual is stream
    # test that setting to existing value returns None
    actual = h.setStream(old)
    assert actual is None


def test_streamhandler_repr():
    class StreamWithName:
        level = picologging.NOTSET
        name = "beyonce"

    handler = picologging.StreamHandler(StreamWithName())
    assert repr(handler) == "<StreamHandler beyonce (NOTSET)>"

    class StreamWithIntName:
        level = picologging.NOTSET
        name = 2

    handler = picologging.StreamHandler(StreamWithIntName())
    assert repr(handler) == "<StreamHandler 2 (NOTSET)>"


def test_streamhandler_handle_return_value():
    handler = picologging.StreamHandler()
    record = picologging.LogRecord(
        "test", picologging.WARNING, __file__, 1, "test", (), None, None, None
    )

    assert handler.handle(record) is True
    assert handler.emit(record) is None

    class TestFilter(picologging.Filter):
        def filter(self, record):
            return False

    handler.addFilter(TestFilter())
    assert handler.handle(record) is None
    assert handler.emit(record) is None



from setuptools import find_packages
from skbuild import setup

with open("./README.md") as fh:
    long_description = fh.read()


setup(
    name="picologging",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    package_data={
        "picologging": ["py.typed", "__init__.pyi", "config.pyi", "handlers.pyi"]
    },
    version="0.9.0",
    author="Microsoft",
    description="A fast and lightweight logging library for Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/microsoft/picologging",
    license="MIT License",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: System :: Logging",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    install_requires=[],
    python_requires=">=3.7",
    extras_require={
        "dev": [
            "rich",
            "pytest",
            "pytest-cov",
            "hypothesis",
            "flaky",
            "black",
            "pre-commit",
        ]
    },
    project_urls={
        "Source": "https://github.com/microsoft/picologging",
        "Documentation": "https://microsoft.github.io/picologging/",
    },
)
